// $Id: Net_Service.C,v 1.2 2006/06/29 01:04:59 orts_mburo Exp $

// (c) Michael Buro 2003, mburo@cs.ualberta.ca, licensed under the GPL

#include "Net_Service.H"
using namespace std;

Net_Service::Net_Service(string& server, int port, string& alias,
                 string& pass,   string& dbFile, string& rootNick_, string& rootPass_, bool verbose_) :
  Net_GGSClient(alias, pass, server, port, true)
{
  
  invitList = new Net_InvitList;
  db = new Net_DB(dbFile);
  
  q_mut = SDL_CreateMutex();
  client_mut = SDL_CreateMutex();

  serve_sem = SDL_CreateSemaphore(0);
  
  rootNick = rootNick_;
  rootPass = rootPass_;
  
  // add root to the database
  DB_Player root(rootNick, rootPass,  ROOT);
  db->put(root);

  verbose = verbose_;
}


Net_Service::~Net_Service() {
  delete invitList;
  delete db;
  SDL_DestroySemaphore(serve_sem);
  SDL_DestroyMutex(client_mut);
  SDL_DestroyMutex(q_mut);
}

/**
   prints the given string if running on verbose mode
*/
void Net_Service::print(const string& out){
  if(verbose){
    time_t t;
    time(&t);
    char ts[27];
    asctime_r(localtime(&t) , ts);
    ts[strlen(ts) - 1] = 0; // delete the new line
    cout << "SERVICE: " << out << " " << ts << endl;
  }
}
  

/**
 * it is run as a thread
 * blocks waiting for a message
 * once a message is recieved it creats a thread of dispatchRequest
 */
sint4 Net_Service:: listen()
{
  string request;
    
  while (!shut_down) {

    this->recv(request);
  
    if (request.size() > 0) {
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION
      if (SDL_mutexP(q_mut)) ERR("mutex 1");
      // put request
      queue.push_front(request);
      if (SDL_mutexV(q_mut)) ERR("mutex 2");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION

      SDL_SemPost(serve_sem); // signal read that there is data available
    }
  }

  return 0;
  //  exit(0); // this thread is done
}


int dispatch(void* data)
{
  // message parsers
  Invite     invite;
  Join       join;
  Leave      leave;
  Ready      ready;
  Unready    unready;
  Kick       kick;
  Abort      abort_;
  Tell       tell;

  Register   register_;
  Login      login;
  List       list;

  Buddy      buddy;
  Worker_Ready wready;

  Down       down;
  
  GGS_Notify notify;
  Finger     finger;

  Error      error;
  
  Worker_Available wavailable;

  Net_Service &service = *static_cast<Net_Service*>(data);

  service.print("[------------DISPATCH------------]");

  string request;
  Net_Message m;
		
  while (!shut_down ) {

    // wait for a request
    sint4 sw;
    while ((sw = SDL_SemWait(service.serve_sem))) {
      REM("NetService dispatch loop awoken prematurely");
    }
     
    // <<<<<<<<<<<<<<<< START CRITICAL SECTION
    if (SDL_mutexP(service.q_mut)) ERR("mutex 1");
    // get request form queue
    if (service.queue.size() <= 0) {
      if (SDL_mutexV(service.q_mut)) ERR("mutex 2");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION
      continue;
    }
    request = service.queue.back();
    service.queue.pop_back();
    if (SDL_mutexV(service.q_mut)) ERR("mutex 2");
    // >>>>>>>>>>>>>>>> END CRITICAL SECTION

    m.setMessage(request);
        
    if (m.getTo() == ".chat") { continue; }
    if (down.parse(m))      { service.serve_down(down);           continue; }
    if (buddy.parse(m))     { service.serve_buddy(buddy);         continue; }
    if (notify.parse(m))    { service.serve_notify(notify);       continue; }
    if (invite.parse(m))    { service.serve_invite(invite);       continue; }
    if (join.parse(m))      { service.serve_join(join);           continue; }
    if (kick.parse(m))      { service.serve_kick(kick);           continue; }
    if (leave.parse(m))     { service.serve_leave(leave);         continue; }
    if (ready.parse(m))     { service.serve_ready(ready);         continue; }
    if (unready.parse(m))   { service.serve_unready(unready);     continue; }
    if (abort_.parse(m))    { service.serve_abort(abort_);        continue; }
    if (register_.parse(m)) { service.serve_register(register_);  continue; }
    if (login.parse(m))     { service.serve_login(login);         continue; }
    if (wready.parse(m))    { service.serve_worker_ready(wready); continue; }
//    if (error.parse(m)) { service.serve_error(error); continue;} 
    if (wavailable.parse(m)){ service.serve_worker_available(wavailable); continue; }

    if (list.parse(m)) {
      if(list.get_list_name()== INVITE_LIST){ service.serve_invitations(list); continue; }
      if(list.get_list_name()== BUDDY_LIST) { service.serve_buddies(list); continue; }
      if(list.get_list_name()== GAME_LIST)  { continue; }
      if(list.get_list_name()== WHO_LIST)   { service.serve_who(list); continue; }
      continue;
    }
    
    if (tell.parse(m)) { service.serve_tell(tell); continue; }

    if (m.getFrom() == "G") { continue; }
    
    if (m.getArgument(0) == "print"){
      service.db->print();
      continue;
    }

    service.print("Message not parsed : " + m.toString());
    
    // invalid
    service.tellClient(m.getFrom(), Error::compose("Invalid command"));
  }

  return 0;
  // exit(0); // this thread is done
}
  
  
/**
 * starts listening for service requests
 * returns -1 if the starting failed
 */
sint4 Net_Service:: start()
{
  FORS (i, MAX_THREADS) {
    if (SDL_CreateThread(dispatch, this) < 0) return -1;
  }
  
  listen();
  return 1;
}

void Net_Service::shutDown()
{
  print("Service is shutting down...");
  this->close();
  db->close();
  //exit(0);
}


bool Net_Service::isOnline(const string& name){
  return !(allPlayers.find(name) == allPlayers.end());
}


/**
   picks a worker, starts the game
*/
void Net_Service::startGame(const Net_Invitation& inv)
{
  print("[------------------i am starting a game ------------------]");

  // generate passes
  Vector<string> ips;
  Vector<string> passes;
  generatePasses(inv, ips, passes);

  if(workerQ.size() > 0) {
    string& workerName =(workerQ.front()).first;
    string& workerHost =(workerQ.front()).second;

    tellClient(workerName, Worker_Start::compose(inv.getOptions(), inv.getPlayerCount(), ips, passes));
    

    const Vector<string> &all = inv.getPlayers();
    const Vector<uint1>  present = inv.getPresent();

    // <<<<<<<<<<<<<<<< START CRITICAL SECTION
    if (SDL_mutexP(client_mut)) ERR("mutex 1");
    string info;

    int j = 0;
    FORU(i, all.size()){
      if(present[i]){
        info = Game_Info::compose(workerHost, passes[j]);
        j++;
        this->send("tell " + all[i] + " " + info);
      }
    }
    
    readyGameQ[workerName] = inv;

    workerQ.pop_front();
    print("*** POPPING from workerQ");
    
  }else{
    print("*** WORKER QUEUE EMPTY! pushing to gameQ");
    gameQ.push_back(Worker_Start::compose(inv.getOptions(), inv.getPlayerCount(), ips, passes)); /** fixme */
  }

}


/**
   generates some random passwords for the player
   @todo: generate random passwords :)
*/
void Net_Service::generatePasses(const Net_Invitation& inv, Vector<string>& ips, Vector<string>& passes){
  int num = inv.getPlayerCount();

  for(int i = 0; i < num; i ++){
    ips.push_back("dummy");
    passes.push_back("dummy");
  }
  
}
 
/*-------------------------------------------- Service Functions -------------------------------*/

/**
   add/remove the given player from senders buddy list
*/
void Net_Service::serve_buddy(Buddy& buddy)
{
  print("[------------BUDDY---------------]");
  
  if (!isOnline(buddy.get_sender())) return;

  if(buddy.get_buddy_name() == buddy.get_sender()){ // cant add your self to your own buddy list 
    tellClient(buddy.get_sender(),Error::compose( "You cannot add yourself as a buddy" ));
    return;
  }
  
  DB_Player me  (buddy.get_sender());
  DB_Player him (buddy.get_buddy_name());
  
  if (db->get(me)  < 0) return;

  if (db->get(him)  < 0){
    tellClient(buddy.get_sender(),Error::compose(buddy.get_buddy_name() + " is not a registered player" ));
    return;
  }
  
  if (buddy.get_add()) { // add
    me.addBuddy(buddy.get_buddy_name());
    tellClient(buddy.get_sender(), Added_Buddy::compose(buddy.get_buddy_name(), isOnline(buddy.get_buddy_name())));
  }else {                // remove
    me.remBuddy(buddy.get_buddy_name());
    tellClient(buddy.get_sender(), Remed_Buddy::compose(buddy.get_buddy_name()));
  }

  db->put(me);
  
  db->print(); //--
}

void Net_Service::serve_buddies(List& list){
  print("[------------BUDDIES-------------]");
  
  if (!isOnline(list.get_sender())) return;

  DB_Player player(list.get_sender());

  if (db->get(player) < 0){
    print("Could not retrieve record of " + list.get_sender() + \
	  "\n\tHow is he/she logged in ?!");
    return;
  }

  Vector<string>* buddyList = player.getBuddies();
  Vector<uint1>   onlineList;
  
  FORALL(*buddyList, i){
    if(isOnline(*i)){
      onlineList.push_back(true);    
    }else{
      onlineList.push_back(false);  
    }
  }
  
  tellClient(list.get_sender(), Buddies::compose(*buddyList, onlineList));
}


/**

*/
void Net_Service::serve_who(List& list){
  print("[------------WHO-----------------]");
  string who;

  if(!isOnline(list.get_sender())){ return; }
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION
  if (SDL_mutexP(client_mut)) ERR("mutex 1");
  FORALL (allPlayers, i){
    who += " ";
    who += (*i).first;
  }
  if (SDL_mutexV(client_mut)) ERR("mutex 2");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION

  tellClient(list.get_sender(), Who_List::compose(who));
}



/**
   if the sender is logged in and has level ROOT
   shut down the service
*/
void Net_Service::serve_down(Down& down)
{
  PlayerList::iterator root;
  root = allPlayers.find(down.get_sender());
  // todo : db->close();
  if (root == allPlayers.end()) return;

  if((*root).second.second == (int) ROOT){
    shutDown();
  }

  signal_handler_shut_down(0);
  signal_handler_exit(0);
}



/**
   creates a game invitation, and notifies
   ( all online players if it is public or
   buddies if it is invisible )
   that an invitation has been created.
   joins the creater to the game
*/

void Net_Service::serve_invite(Invite& invite)
{
  print("[------------INVITE--------------]");
  // authenticate : player is logged in
  if (!isOnline(invite.get_sender())) return;
  
  string name = invite.get_sender();

  Net_Invitation *n = new Net_Invitation(invite);
    
  // "alias" may be null because when someone created the game and selected 
  // "anonymized ids" (creategame.ui.h) they did not enter an alias. In this case 
  // we will use their username.
  string alias;
  if( invite.get_alias() != "null" )alias = invite.get_alias();
  else alias = invite.get_sender();  
  
  sint4 pid = n->addPlayer( invite.get_sender(), alias );
  invitList->add(n);
  fprintf(stderr, "___*** Invite Created with ID: %d\n", n->getIID());
  
  tellInvitation( *n, Joined::compose(n->getIID(), pid, n->getAnony(), alias) );
  tellClient(invite.get_sender(), Player_List::compose(n->getPresent(), n->getAliases(), n->getReady()));

  if(n->getInvisible()){
    DB_Player player(invite.get_sender());
    
    if(db->get(player)>0){
      tellBuddies(player, Invitations::compose(n->getInfo()));
    }else{
      ERR("Client " + invite.get_sender() + " is online but not in the database !!!");
    }

  }else{
    tellAll(Invitations::compose(n->getInfo()));
  }
  
  invitList->print();
}




/**
   checks if the player is the owner
   kills the game
   notifies all members that it has been killed
*/

void Net_Service::serve_abort(Abort& abort)
{
  // authenticate : player is the owner
  //                player is logged in
  print("[------------ABORT---------------]");
  
  if (!isOnline(abort.get_sender())) return;
  
  Net_Invitation* inv = invitList->get((uint4)abort.get_iid());
  if (!inv){
    char s[256];
    sprintf(s,"Cannot abort:\n\tGame with id %d could not be found",abort.get_iid());
    tellClient(abort.get_sender(),Error::compose(string(s)));
    return;
  }

  if (abort.get_sender() == inv->getOwner()) {
    invitList->del(abort.get_iid());
    fprintf(stderr, "___*** Invite Deleted with ID: %d\n", abort.get_iid());
    tellAll(Aborted::compose(abort.get_iid()));
  }else{
    char s[256];
    sprintf(s,"Cannot abort:\n\tYou are not the owner of game %d",abort.get_iid());
    tellClient(abort.get_sender(),Error::compose(string(s)));
  }

  invitList->print(); //--
}





/**
   checks if the player is the owner
   kicks the reqested member out
   notifies all members
*/

void Net_Service::serve_kick(Kick& kick)
{
  print("[------------KICK----------------]");
  // authenticate : player is the owner
  //                player is loged in

  if (kick.get_pid() == 0) return; // cant kick yourself... i donno why, just dont like it

  if (!isOnline(kick.get_sender())) return;

  Net_Invitation* inv = invitList->get((uint4)kick.get_iid());

  if (inv == NULL) {
    stringstream msg;
    msg << "invitation with iid " << kick.get_iid() << " not found";
    print(msg.str());
    tellClient(kick.get_sender(), Error::compose(msg.str()));
    return;
  }

  if (kick.get_sender() == inv->getOwner()) {
    string msg = kick.get_sender() + " is not the owner\n\tthe owner is " + inv->getOwner();
    print(msg);
    tellClient(kick.get_sender(), Error::compose(msg));
    
    tellInvitation(*inv, Kicked::compose(inv->getIID(), kick.get_pid()));
    inv->removePlayer(kick.get_pid());
  }
  inv->print();
}




/**
   checks if player is registered
   adds player to the game
   notifies all members
*/

void Net_Service::serve_join(Join& join)
{
  print("[------------JOIN----------------]");
  // authenticate : player is logged in
  string name = join.get_sender();

  if (!isOnline(join.get_sender())) return;

  Net_Invitation* inv = invitList->get((uint4)join.get_iid());
  fprintf(stderr, "___*** Join ID: %d\n", join.get_iid());
  
  if (inv) {
    if( join.get_passprotected() ){
      if(join.get_pass() != inv->getPass() /*&& join.get_passprotected()*/ ){
	tellClient(join.get_sender(), Error::compose("Incorrect password for this game"));
	cout << "reallPass |" << inv->getPass()  << "|" << endl;
	cout << "givenPass |" << join.get_pass() << "|" << endl;
	return;
      }
    }
    

    // "alias" may be null because they did not enter an anonymizing id 
    // when prompted to in AnonymizingDialog (anonymizingdialog.ui.h).
    // In this case we will use their username.
    string alias;
    if( join.get_alias() != "null") alias = join.get_alias();
    else alias = join.get_sender();

    int pid = inv->addPlayer(join.get_sender(), alias);
    if (pid > 0) {
      
      if( join.get_anony_ids() ){
	
	
	tellClient(join.get_sender(), Player_List::compose(inv->getPresent(), inv->getAliases(), inv->getReady()));
      }
      else{ 
	alias = join.get_sender(); 
	tellClient(join.get_sender(), Player_List::compose(inv->getPresent(), inv->getPlayers(), inv->getReady()));
      
      }
      tellInvitation( *inv, Joined::compose(inv->getIID(), pid, inv->getAnony(), alias) );
    }
  }

  if (inv) inv->print();
}





/**
   if player is in the game
   removes player
   notifies all members
*/

void Net_Service::serve_leave(Leave& leave)
{
  print("[------------LEAVE---------------]");
  // authenticate : the sender must match the player alias
  //                player is logged in

  if (!isOnline(leave.get_sender())) return;

  Net_Invitation* inv = invitList->get((uint4)leave.get_iid());

  if (inv) {
    
    sint4 pid = inv->removePlayer(leave.get_sender());

    if(pid < 0) {
      char s[256];
      sprintf(s,"You are not a member of game %d", leave.get_iid());
      tellClient(leave.get_sender(), Error::compose(string(s)));
      return;
    }

    tellInvitation(*inv, Left::compose(inv->getIID(), pid));
    tellClient(leave.get_sender(), Left::compose(inv->getIID(), pid));
    
    if (inv->getPlayerCount() == 0 || pid == 0) { // everybody gone :(... kill game
      invitList->del(leave.get_iid());
      fprintf(stderr, "___*** Invite Deleted with ID: %d\n", leave.get_iid());
      tellAll(Aborted::compose(leave.get_iid()));
      return;
    }
    inv->print(); //--
  }
}

/**
   if player is in the game
   player status is readied
   all players are notified
*/
void Net_Service::serve_ready(Ready& ready)
{
  print("[------------READY---------------]");
  //authenticate : player is logged in
  //               player is in game

  if (!isOnline(ready.get_sender())) return;
  
  Net_Invitation* inv = invitList->get((uint4)ready.get_iid());

  if (inv) {
    sint4 alias = inv->setReady(ready.get_sender(), true);
    if(alias < 0){
      stringstream msg;
      msg << "not currently in game with iid " << ready.get_iid();
      print( "Player " + ready.get_sender() + " is " + msg.str());
      tellClient(ready.get_sender(), Error::compose("You are " + msg.str()));
      return;
    }
    tellInvitation(*inv, Readied::compose(inv->getIID(), alias));
    if (inv->allReady()) {
      startGame(*inv);
    }
  }

  if(inv) inv->print(); //--
}




/**
   if player is in the game
   player status is unreadied
   all players are notified
*/

void Net_Service::serve_unready(Unready& unready)
{
  print("[------------UNREADY-------------]");
  // authenticate : palyer is in game
  //                player is logged in

  if (!isOnline(unready.get_sender())) return;

  Net_Invitation* inv = invitList->get((uint4)unready.get_iid());
  if (inv){
    sint4 alias = inv->setReady(unready.get_sender(), false);
    tellInvitation(*inv, Unreadied::compose(inv->getIID(), alias));
  }
  if(inv) inv->print(); //--
}


void Net_Service::serve_notify(GGS_Notify& notify)
{
  print("[------------NOTIFY--------------]");
  if (!notify.get_online()) {

    DB_Player player(notify.get_name());

    if (db->get(player) < 0) return;

    // tell buddies player went offline
    tellAll(Login_Notify::compose(false, notify.get_name()));

    // remove player from online list
    allPlayers.erase(notify.get_name());
    
    // remove player from all invitations
    Net_Invitation* inv;
    
    for(uint4 i = 0; i < invitList->getSize(); i++){
      inv = invitList->get(i);
      if(inv){
      
        int pid = inv->removePlayer(notify.get_name());

        if(pid < 0){
          return;
        }
        
        tellInvitation(*inv, Left::compose(i, pid));
        inv->print();

        if (inv->getPlayerCount() == 0 || pid == 0) { // everybody gone :(... kill game
          invitList->del(i);
          fprintf(stderr, "___*** Invite Deleted with ID: %d\n", i);
          tellAll(Aborted::compose(i));
        }
      }
    }
  }
}




/**
   root can register super
   super can register general
   check user (name, pass, level)
   add the new registration to the database
*/

void Net_Service::serve_register(Register& reg)
{
  print("[------------REGISTER------------]");
  // add player to the database

  // case 1 : new entry
  //          level = general || worker        open to everybody
  //          level = super                    open to root only

  // case 2 : existing entry (changing password)
  //          level = general || worker        open to super,root, and the record owner
  //          level = super                    open to super,and root
  //          level = root                     open to root

  // retrieve record
  DB_Player player(reg.get_name());

  bool exists = (db->get(player) > 0);
  bool allow  = false;


  PlayerList::iterator i = allPlayers.find(reg.get_sender());

  if(i != allPlayers.end() && ((*i).second).second == ROOT){
    allow = true;
  }
  
  if(i == allPlayers.end() && exists){
    tellClient(reg.get_sender(), Error::compose("User name " + reg.get_name() + " already exists"));
    return;
  }

  if(exists && !allow){
    
    if( i != allPlayers.end() ){
      // if the entry exists allow it if
      //    level is not changing && sender is the entry owner
      //    sender has a higher level && requested level is < sender

      if(reg.get_level() == player.getLevel()){     // level not changing
	allow = true;
      }else{
	tellClient(reg.get_sender(), Error::compose("You cannot change your registration level"));
	return;
      }

      if( !((((*i).second).second >  reg.get_level()) || // u have a higher level
	  (reg.get_sender() == player.getName()   )) ){ // u are the owner
	tellClient(reg.get_sender(), Error::compose("You cannot change password unless you are the owner of the record or you have a higher level"));
	return;
      }
    }
      
  }else{
    if(!allow){
    // if the entry does not exist allow it if the user wants to register
    // as worker or general
      allow = (reg.get_level() == WORKER || reg.get_level() ==  GENERAL);
    }
  }

  if(!allow){ 
    tellClient(reg.get_sender(), Error::compose("This is an illegal request."));
    return;   
  }

  /* vvvvvvvvvvvvvvvvvvvvv request is legal now add the entry vvvvvvvvvvvvvvvvvvvvvvvvvv */

  if(reg.get_level() ==  WORKER){
    DB_Player worker(reg.get_name(), reg.get_pass(),  WORKER);
    db->put(worker);
  }else{
    if(exists){
      player.setPass(reg.get_pass());
      db->put(player);
    }else{
      DB_Player entry(reg.get_name(), reg.get_pass(), (Level_)reg.get_level());
      db->put(entry);
    } 
  }
  
}



/**
   authenticate the player
   (nick, pass, level)
   add them to the online players list
   tell buddies that player is online ---- fixme
   ask GGS to notify the service when this player goes offline
*/

void Net_Service::serve_login(Login& log)
{
  print("[------------LOGIN---------------]");
  
  if (isOnline(log.get_sender())){
    // player is already logged in
    return;
  }

  // check password
  DB_Player player(log.get_sender());

  bool found =
    db->get(player) > 0                   &&
    log.get_sender()  == player.getName() &&
    log.get_pass()  == player.getPass()   &&
    log.get_level() == player.getLevel();

  if(found){
    print(log.get_sender() + " logged in successfuly");
  }else{
    print(log.get_sender() + " failed to login");
  }

  if(found) {
    if(log.get_level() ==  WORKER){

      // notfy worker that the login was success ful
      tellClient(log.get_sender(), Login_Notify::compose(true, log.get_sender()));
      
      // start any waiting games
      if (gameQ.size() > 0) {
        print("*** POPPING from gameQ");
	tellClient(log.get_sender(), gameQ.front());
        gameQ.pop_front();
      } else {
        workerQ.push_back(make_pair(log.get_sender(), log.get_ip()));
	print("*** GAME QUEUE EMPTY! pushing to workerQ");
      }

    } else {
      allPlayers[log.get_sender()] = make_pair(log.get_sender(), log.get_level());
      // tellBuddies that player is online
      tellAll(Login_Notify::compose(true, log.get_sender()));
            
      // ask GGS to notify service when player goes off line
      tellGGS(GGS_Notify::compose(log.get_sender(), true));
      
    }
  }else{ // not found
    tellClient(log.get_sender(), Error::compose("Login failed.\nInvalid information."));
  }
}

void Net_Service::serve_invitations(List& list){
  Net_Invitation* current;
  FORU(i, invitList->getSize()){
    current=invitList->get(i);
    if (current!=0) {
      if (!current->getInvisible()) { //fix so if on buddy list send anyways
        tellClient(list.get_sender(), Invitations::compose(current->getInfo()));
      }
    }
  }
}

void Net_Service::serve_tell(Tell& tell){
  print("[------------TELL----------------]");
  Net_Invitation* inv = invitList->get(tell.get_iid());
  if (!inv){
    stringstream msg;
    msg << "invitation with iid " << tell.get_iid() << " not found" << endl;
    print(msg.str());
    tellClient(tell.get_sender(), Error::compose(msg.str()));
    return;
  }
  tellInvitation(*inv, Tell::compose(tell.get_iid(), tell.get_pid(), tell.get_out_msg()));
}

void Net_Service::serve_worker_ready(Worker_Ready& wready){
  print("[------------WREADY--------------]");
  ReadyGameQ::iterator i;
  
  i = readyGameQ.find(wready.get_sender());

  if(i == readyGameQ.end()){
    tellClient(wready.get_sender(),Error::compose("Worker_Ready message invalid: No pending game for you"));
    return;
  }
  
  tellInvitation((*i).second, Game_Start::compose());
}

void Net_Service::serve_worker_available(Worker_Available& wavailable) {
  
  print("[------------WAVAILABLE----------]");
  
  workerQ.push_back(make_pair(wavailable.get_sender(), wavailable.get_ip()));
  print("*** Pushing to workerQ, IP: " + wavailable.get_ip());
  
}

/*--------------------------------------- tell funcitons -------------------------------------------- */

void Net_Service::tellInvitation(const Net_Invitation& inv, const string& message)
{
  const Vector<string> &all = inv.getPlayers();
  const Vector<uint1>  present = inv.getPresent();

  // <<<<<<<<<<<<<<<< START CRITICAL SECTION
  if (SDL_mutexP(client_mut)) ERR("mutex 1");
  FORU(i, all.size()){
    if(present[i]){
      this->send("tell " + all[i] + " " + message);
    }
  }
  if (SDL_mutexV(client_mut)) ERR("mutex 2");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION
}

void Net_Service::tellAll(const string& message)
{
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION
  if (SDL_mutexP(client_mut)) ERR("mutex 1");
  FORALL (allPlayers, i) this->send("tell " + (*i).first + " " + message);
  if (SDL_mutexV(client_mut)) ERR("mutex 2");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION
}

void Net_Service::tellBuddies(DB_Player& player, const string& message)
{
  const Vector<string> &buddies = *player.getBuddies();

  // <<<<<<<<<<<<<<<< START CRITICAL SECTION
  if (SDL_mutexP(client_mut)) ERR("mutex 1");
  FORALL (buddies, i) this->send("tell " + *i + " " + message);
  if (SDL_mutexV(client_mut)) ERR("mutex 2");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION
}



