/** @file GfxEmitter.C
    @see GfxEmitter.H

    $Id: GfxEmitter.C,v 1.44 2006/08/11 20:54:01 dtom Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/gfxclient/src/GfxEmitter.C,v $
*/

// This is an ORTS file
// (c) Keith Yerex
// (c) Tim Furtak
// (c) Erik Benner
// licensed under the GPL


#include "GfxEmitter.H"
#include "GfxModule.H"
#include "Image.H"

using namespace std;

//===================================================================

void GfxEmitter::beginDraw()
{
  glEnable(GL_BLEND);
  glDepthMask(GL_FALSE);
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);

  glActiveTextureARB(GL_TEXTURE0);
  glEnable(GL_TEXTURE_2D);
  
  glActiveTextureARB(GL_TEXTURE1);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_TEXTURE_GEN_S);
  glDisable(GL_TEXTURE_GEN_T);

  glColor4f(1,1,1,1);
}

//-------------------------------------------------------------------

void GfxEmitter::endDraw()
{
  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);

  glActiveTextureARB(GL_TEXTURE0);
}

//===================================================================

void Particle::update(real4 dt)
{
  pos += vel*dt+(acc+vel*fric)*dt*dt*0.5;
  vel += (acc+vel*fric)*dt;
  age += dt;
}

// Pass the global acceleration instead of storing it per particle
bool ParticleLite::update(real4 dt, const Vec3<real4> &acc)
{
  pos += vel*dt;
  vel += acc*dt;
  age += dt;
  if (age >= life) {
    life = 0.0f;
    return false;
  } else
    return true;
}

//===================================================================

GfxPListEmitter::GfxPListEmitter()
{
  first = last = 0;
}

//-------------------------------------------------------------------

void GfxPListEmitter::reorder()
{
  sint4 nmax = int(ceil(rate) * ceil(life));
  if (nmax <= max_particles) return;

  Vector<Particle> old_p;

  for (int i = first; i != last; i = i % max_particles) {
    int e = i < last ? last : max_particles;
    for (; i < e; ++i) {
      old_p.push_back(list[i]);
    }
  }

  list.resize(nmax);
  assert(list.size() >= old_p.size());
  FORU (i, old_p.size()) {
    list[i] = old_p[i];
  }
  first = 0;
  last = old_p.size() % nmax;
  if (last == first && old_p.size() != 0) first = (first+1) % nmax;
  max_particles = nmax;
}

//-------------------------------------------------------------------

void GfxPListEmitter::set_rate(real4 r)
{
  rate = r;
  reorder();
}

//-------------------------------------------------------------------

Particle& GfxPListEmitter::newparticle()
{
  Particle *p = &list[last];
  last = (last+1) % max_particles;
  if (last == first) first = (first+1) % max_particles;
  return *p;
}

//===================================================================

GfxBoxEmitter::GfxBoxEmitter(GfxModule &gfxm_, real4 rate_, real4 life_, real4 time_limit,
                             Vec3<real4> p, Vec3<real4> pr, Vec3<real4> v, Vec3<real4> vr, Vec3<real4> accel,
                             real4 f, real4 fr, real4 p_width, real4 p_height,
                             GLuint texid, GLuint colid)
{
  set_gfxm(gfxm_);
  dying = false;
  rate = rate_;
  life = life_;
  limit = time_limit;
  if (rate < 0) rate = 0;
  if (life < 0) life = 0;
  max_particles = int(ceil(rate) * ceil(life));

  if (max_particles < 1) max_particles = 1;

  pos = p;
  p_rand = pr;
  vel = v;
  v_rand = vr;
  acc = accel;

  fric = f;
  f_rand = fr;

  w = p_width;
  h = p_height;
  billboard_tex = texid;
  color_tex = colid;
  blend = STANDARD;

  pose.center = Vec3<real4>(0, 0, 0);

  list.resize(max_particles);

  unlimited = limit <= 0; // initial setting

  create = 0;
  enabled = true;

  radius = life * (vel + v_rand).length();

  glBindTexture(GL_TEXTURE_2D, colid);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
}

//-------------------------------------------------------------------

void GfxBoxEmitter::spawn_particle()
{
  Particle &p = newparticle();

  p.pos = pose.center + pos + Vec3<real4>::vrand(gfxm->rand)*p_rand;

#if 1
  // adjust particle velocity based on object rotation
  Vec3<real4> v = vel + Vec3<real4>::vrand(gfxm->rand) * v_rand;
  p.vel.x = (v.x * cos_ang) + (v.z * sin_ang);
  p.vel.y = v.y;
  p.vel.z = (v.x * -sin_ang) + (v.z * cos_ang);
#else
  p.vel = vel + Vec3<real4>::vrand(gfxm->rand) * v_rand;
#endif
  p.acc = acc;
  p.fric = fric + gfxm->rand.rand_real8() * f_rand - f_rand / 2;
  p.age = 0.0;
  p.life = life;
}

//-------------------------------------------------------------------

void GfxBoxEmitter::update(real4 dt)
{
  GfxObject::update(dt);

  if (first == last && dying == true) {
    dead = true;
    return;
  }

  aabb_o.x = pose.center.x - radius;
  aabb_o.y = pose.center.y - radius;
  aabb_o.z = pose.center.z - radius;
  aabb_d.x = aabb_d.y = aabb_d.z = 2*radius;
  
  for (int i = first; i != last; i = i % max_particles) {
    int e = i < last ? last : max_particles;
    for (; i < e; ++i) {
      list[i].update(dt);
      // assumes all particles live the same amount of time
      if (list[i].age >= list[i].life) {
        first = (i+1) % max_particles;
      }
    }
  }

  if ((unlimited || limit>0) && enabled) {
    if (limit > 0) limit -= dt;

    double ang = pose.y_rot * M_PI / 180.0;
    cos_ang = cos(ang);
    sin_ang = sin(ang);
    create += rate*dt;
    while (create >= 1) {
      spawn_particle();
      create--;
    }
  }
}

//===================================================================

GfxSprayEmitter::GfxSprayEmitter(GfxModule &gfxm, real4 rate_, real4 life_, real4 time_limit,
                                 const Vec3<real4> &p, real4 min_lat, real4 max_lat,
                                 real4 offset_, real4 o_rand_, real4 speed_, real4 s_rand_,
                                 const Vec3<real4> &accel, real4 f, real4 fr,
                                 real4 ww, real4 hh, GLuint texid, GLuint colid)
{
  set_gfxm(gfxm);
  dying = false;
  rate = rate_;
  life = life_;
  limit = time_limit;
  if (rate < 0) rate = 0;
  if (life < 0) life = 0;
  max_particles = int(ceil(rate) * ceil(life));

  if (max_particles < 1) max_particles = 1;

  pos = p;
  min_ang = min_lat * M_PI / 180.0;
  max_ang = max_lat * M_PI / 180.0;
  offset = offset_;
  o_rand = o_rand_;
  speed = speed_;
  s_rand = s_rand_;

  acc = accel;

  fric = f;
  f_rand = fr;

  w = ww;
  h = hh;
  billboard_tex = texid;
  color_tex = colid;
  blend = STANDARD;

  pose.center = Vec3<real4>(0, 0, 0);

  list.resize(max_particles);

  unlimited = limit <= 0; // initial setting

  create = 0;
  enabled = true;

  radius = life * (speed + s_rand);

  glBindTexture(GL_TEXTURE_2D, colid);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
}

//-------------------------------------------------------------------

void GfxSprayEmitter::spawn_particle()
{
  Particle &p = newparticle();

  // adjust particle velocity based on object rotation
  real8 lat = gfxm->rand.rand_real8() * (max_ang - min_ang) + min_ang;
  real8 theta = gfxm->rand.rand_real8() * 2 * M_PI;
  real8 s = speed + gfxm->rand.rand_real8() * s_rand;
  real8 d = offset + gfxm->rand.rand_real8() * o_rand;

  p.pos = pose.center + pos;
  p.pos.x += cos(lat) * cos(theta) * d;
  p.pos.y += sin(lat) * d;
  p.pos.z += cos(lat) * sin(theta) * d;

  p.vel.x = cos(lat) * cos(theta) * s;
  p.vel.y = sin(lat) * s;
  p.vel.z = cos(lat) * sin(theta) * s;

  p.acc = acc;
  p.fric = fric + gfxm->rand.rand_real8() * f_rand - f_rand / 2;
  p.age = 0.0;
  p.life = life;
}

//-------------------------------------------------------------------

void GfxSprayEmitter::update(real4 dt)
{
  GfxObject::update(dt);

  if (first == last && dying == true) {
    dead = true;
    return;
  }

  aabb_o.x = pose.center.x - radius;
  aabb_o.y = pose.center.y - radius;
  aabb_o.z = pose.center.z - radius;
  aabb_d.x = aabb_d.y = aabb_d.z = 2*radius;
  
  for (int i = first; i != last; i = i % max_particles) {
    int e = i < last ? last : max_particles;
    for (; i < e; ++i) {
      list[i].update(dt);
      // assumes all particles live the same amount of time
      if (list[i].age >= list[i].life) {
        first = (i+1) % max_particles;
      }
    }
  }

  if ((unlimited || limit>0) && enabled) {
    if (limit > 0) limit -= dt;

    create += rate*dt;
    while (create >= 1) {
      spawn_particle();
      create--;
    }
  }
}

//===================================================================

GfxCircleEmitter::GfxCircleEmitter(GfxModule &gfxm, real4 rate_, real4 life_, real4 time_limit,
                                 const Vec3<real4> &p, real4 min_lat, real4 max_lat,
                                 real4 offset_, real4 o_rand_, real4 speed_, real4 s_rand_,
                                 const Vec3<real4> &accel, real4 f, real4 fr,
                                 real4 ww, real4 hh, GLuint texid, GLuint colid)
{
  set_gfxm(gfxm);
  dying = false;
  rate = rate_;
  life = life_;
  limit = time_limit;
  if (rate < 0) rate = 0;
  if (life < 0) life = 0;
  max_particles = int(ceil(rate) * ceil(life));

  if (max_particles < 1) max_particles = 1;

  pos = p;
  min_ang = min_lat * M_PI / 180.0;
  max_ang = max_lat * M_PI / 180.0;
  offset = offset_;
  o_rand = o_rand_;
  speed = speed_;
  s_rand = s_rand_;

  acc = accel;

  fric = f;
  f_rand = fr;

  w = ww;
  h = hh;
  billboard_tex = texid;
  color_tex = colid;
  blend = STANDARD;

  pose.center = Vec3<real4>(0, 0, 0);

  list.resize(max_particles);

  unlimited = limit <= 0; // initial setting

  create = 0;
  enabled = true;

  radius = life * (speed + s_rand);
  rot_amt = 0;

  glBindTexture(GL_TEXTURE_2D, colid);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
}

//-------------------------------------------------------------------

void GfxCircleEmitter::spawn_particle()
{
  Particle &p = newparticle();

  // adjust particle velocity based on object rotation
  real8 lat = gfxm->rand.rand_real8() * (max_ang - min_ang) + min_ang;
  real8 theta = gfxm->rand.rand_real8() * 2 * M_PI;
  real8 s = speed + gfxm->rand.rand_real8() * s_rand;
  real8 d = offset + gfxm->rand.rand_real8() * o_rand;

  p.pos = pose.center + pos;
  p.pos.x += cos(lat) * cos(theta) * d;
  p.pos.y += sin(lat) * d;
  p.pos.z += cos(lat) * sin(theta) * d;

  p.vel.x = cos(lat) * cos(theta) * s;
  p.vel.y = sin(lat) * s;
  p.vel.z = cos(lat) * sin(theta) * s;

  p.acc = acc;
  p.fric = fric + gfxm->rand.rand_real8() * f_rand - f_rand / 2;
  p.age = 0.0;
  p.life = life;
}

//-------------------------------------------------------------------

void GfxCircleEmitter::update(real4 dt)
{
  GfxObject::update(dt);
  rot_amt += 5;
  if (first == last && dying == true) {
    dead = true;
    return;
  }

  aabb_o.x = pose.center.x - radius;
  aabb_o.y = pose.center.y - radius;
  aabb_o.z = pose.center.z - radius;
  aabb_d.x = aabb_d.y = aabb_d.z = 2*radius;
  
  for (int i = first; i != last; i = i % max_particles) {
    int e = i < last ? last : max_particles;
    for (; i < e; ++i) {
      list[i].update(dt);
      // assumes all particles live the same amount of time
      if (list[i].age >= list[i].life) {
        first = (i+1) % max_particles;
      }
    }
  }

  if ((unlimited || limit>0) && enabled) {
    if (limit > 0) limit -= dt;

    create += rate*dt;
    while (create >= 1) {
      spawn_particle();
      create--;
    }
  }
}

//-------------------------------------------------------------------

void GfxCircleEmitter::draw_emitters()
{
  glPushMatrix();

  glScalef(pose.scale, pose.scale, pose.scale);

  if (draw_type == STREAKS) draw_as_streaks();
  else draw_as_points();

  glTranslatef(pose.center.x, pose.center.y, pose.center.z);
  glRotatef(pose.y_rot+rot_amt, 0, 1, 0);
  //  glRotatef(rot_amt, 0, 1, 0);

  //  FORALL (sub_objects, i) (*i)->draw_emitters();
  FORALL (emitters, i) (*i)->draw_emitters();

  glPopMatrix();
}

//===================================================================

void GfxPListEmitter::draw_emitters()
{
  glPushMatrix();

  glScalef(pose.scale, pose.scale, pose.scale);

  if (draw_type == STREAKS) draw_as_streaks();
  else draw_as_points();

  glTranslatef(pose.center.x, pose.c