/** @file Motion.C
    @see Game.H

    $Id: Motion.C,v 1.40 2007/04/02 23:24:52 marcar Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Motion.C,v $
*/

// This is an ORTS file
// (c) Michael Buro
// (c) Tim Furtak
// licensed under the GPL

/*
  Collision categories:

  (ON_SURFACE = ON_LAND || ON_WATER)

  UNDER_WATER: collision with WATER/LAND boundary + FRINGE
  ON_WATER:    collision with ON_WATER objects and WATER/LAND boundary + FRINGE
  ON_LAND:     collision with ON_LAND objects and WATER/LAND and GROUND +
               FRINGE boundaries
  IN_AIR:      collisions with FRINGE (and IN_AIR objects)


  simplified:

  UNDER_WATER: collision with all UNDER_WATER objects and all boundaries
  ON_WATER:    collision with all ON_WATER objects and all boundaries
  ON_LAND:     collision with all ON_LAND objects and all boundaries
  IN_AIR:      collisions with FRINGE (and IN_AIR objects)

*/

#include "Game.H"
#include "GameObj.H"
#include "ServerObjData.H"
#include "GUI.H"
#include "PairHeap.H"
#include "CycleCounter.H"
#include "Array.H"
#include "ServerObjData.H"
#include "GameTile.H"
#include "Options.H"
#include "Profiler.H"

using namespace std;


#define TEST_OUT 0
#define TEST_OBJ 0
#define ZCAT_CHECK 0

#define TOO_CLOSE(x, a, b)                                              \
      {                                                                 \
        cout << "TOO CLOSE " << x << " " << a << " " << b << endl;      \
 /* fixme: no direct gui access anymore  if (gui) gui->display(a, b); FOREVER{}; exit(1);    */            \
      }

//===================================================================

void MotionSectors::init(sint4 num_sec)
{
  assert(num_sec > 0);
  num = num_sec;

  MotionSectorInfo si;
  FORS (i, num) {
    secs.push_back(si);
    sizes.push_back(0);
  }

  if (active) delete [] active;
  active = new uint1[num];
}

//-------------------------------------------------------------------

void MotionSectors::reset_sizes() {
  // cout << secs.size() << " " << sizes.size() << endl;
  assert(secs.size() == sizes.size());
  FORU (i, secs.size()) secs[i].set_size(sizes[i]);
}

//-------------------------------------------------------------------

void MotionSectors::clear() {
  secs.clear(); sizes.clear();
}

//-------------------------------------------------------------------

void MotionSectors::insert(GameObj *g, sint4 s, bool is_static)
{
  assert(s >= 0 && s < num);
  if (!is_static && !active[s]) {
    active[s] = true;
    active_sectors.push_back(s);
  }
  
  MotionSectorInfo &ms = secs[s];

  // The motion sectors could have non-boundary objects in them.
  // The boundary needs to be inserted (anywhere) before those, with the other static objects.

  if (is_static) {
    ms.push_back(g);
    sint4 i = sizes[s]; // original number of boundary objects
    swap(ms[i], ms[ms.size()-1]);
    ++sizes[s];
  } else {
    ms.push_back(g); // fixme: check if duplicates will cause a problem
  }
}

//===================================================================

/** Compute connected component by following neighborhood pointers.
    Store pointers to game objects in pointer array pointed to by comp.
    The resulting network can be checked for moving objects by
    Game::move_objects(Array<GameObj*> &mobjs, MotionSectors &ms, bool obj_coll).
*/

void Game::visit(GameObj &e, GameObj** &comp)
{
#if TEST_OBJ
  if (!e.test_magic_ok()) {
    ERR("oops");
  }
#endif

  if (e.smd.visit_magic == visit_magic) return; // already visited

  e.smd.visit_magic = visit_magic;
  *comp++ = &e;

  const Neighbors &ne = e.smd.neighbors;

  FORS (ni, ne.size()) visit(*ne[ni], comp);
}

//===================================================================

/** Create a boundary object, add it to the relevant movement layers.
    Can be called after motion sector initialization has already taken place;
    places the boundary with the other static objects at the beginning of
    the list and updates the sector size.

    @todo Compute the motion sectors a boundary is added to with a line drawing
          algorithm instead of a rectangle.  Same for add_tile_boundary, and
          remove_tile_boundary.

    @param x  x position of line endpoint 1, given in tile units.
    @param y  y position of line endpoint 1, given in tile units.
    @param x2 x position of line endpoint 2, given in tile units.
    @param y2 y position of line endpoint 2, given in tile units.

    @param is_fringe Whether the boundary is part of the game field hard limits.
*/

GameObj* Game::add_tile_boundary(sint4 x, sint4 y, sint4 x2, sint4 y2,
                                 Tile::Type t1, Tile::Type t2, bool height_diff, bool is_fringe)
{
  // not really needed, but it keeps the line's bounding circle small
  assert(x-x2 <= 1 && x-x2 >= -1);
  assert(y-y2 <= 1 && y-y2 >= -1);

  ScriptObj *temp = NPC->gof.create("boundary");
  GameObj *gob = temp->get_GameObj();
  assert(gob);
  if (!gob) ERR("out of memory");
  gob->init_sod();

  sint4 xp  = x  * tile_points;
  sint4 yp  = y  * tile_points;
  sint4 x2p = x2 * tile_points;
  sint4 y2p = y2 * tile_points;

  *gob->sod.x1 = xp;
  *gob->sod.y1 = yp;
  *gob->sod.x2 = x2p;
  *gob->sod.y2 = y2p;
  gob->sod.is_boundary = true;

  *gob->sod.shape = Object::LINE;

  if (x == x2 || y == y2) {
    sint4 rad = ::max(x2p - xp, y2p - yp);
    gob->smd.bc_fine_r = fine_grid(rad)/2;
  } else {
    gob->smd.bc_fine_r = Scalar(fine_grid(tile_points) * 0.71) +1;
  }

  gob->smd.bc_fine_center.x = fine_grid(*gob->sod.x1 + *gob->sod.x2)/2;
  gob->smd.bc_fine_center.y = fine_grid(*gob->sod.y1 + *gob->sod.y2)/2;

  *gob->sod.ncollides = 0;
  if (is_fringe) {
    *gob->sod.collides = (uint4)(-1);
    *gob->sod.zcat = Object::ZCAT_UNDEF;
  } else {
    if (height_diff) {
      *gob->sod.collides = tile_height_diff_mask;
    } else {
      *gob->sod.collides = tile_collision_mask[num_tile_types * t1 + t2];
    }
    *gob->sod.zcat = Object::ON_LAND;
  }

  boundaries.push_back(gob);

  // compute sectors

  sint4 sx1 = fine_grid(xp)/sec_wh;
  sint4 sy1 = fine_grid(yp)/sec_wh;
  sint4 sx2 = fine_grid(x2p)/sec_wh;
  sint4 sy2 = fine_grid(y2p)/sec_wh;

  // take the rectangle of motion sectors and mark them
  // a line drawing algorithm would be better

  if (sx2 < sx1) swap(sx1,sx2);
  if (sy2 < sy1) swap(sy1,sy2);

  if (sx1 < 0) {
    sx1 = 0;
    if (sx2 < 0) sx2 = 0;
  }
  if (sx2 >= sec_nx) {
    sx2 = sec_nx-1;
    if (sx1 >= sec_nx) sx1 = sec_nx-1;
  }

  if (sy1 < 0) {
    sy1 = 0;
    if (sy2 < 0) sy2 = 0;
  }
  if (sy2 >= sec_ny) {
    sy2 = sec_ny-1;
    if (sy1 >= sec_ny) sy1 = sec_ny-1;
  }

  for (sint4 i = sx1; i <= sx2; ++i) {
    for (sint4 j = sy1; j <= sy2; ++j) {
      sint4 ind = j * sec_nx + i;

      if (is_fringe) {
        fringe_secs.secs[ind].push_back(gob);
      }

      boundary_secs.secs[ind].push_back(gob);
      boundary_secs.sizes[ind]++;

      motion_secs.insert(gob, ind, true);
    }
  }

  return gob;
}

//===================================================================

/** Remove a boundary aligned on the tile grid.
    Maintains the validity of the motion sectors affected by the removal.

    @param x  x position of line endpoint 1, given in tile units.
    @param y  y position of line endpoint 1, given in tile units.
    @param x2 x position of line endpoint 2, given in tile units.
    @param y2 y position of line endpoint 2, given in tile units.
*/

void Game::remove_tile_boundary(sint4 x, sint4 y, sint4 x2, sint4 y2)
{
  sint4 xp  = x  * tile_points;
  sint4 yp  = y  * tile_points;
  sint4 x2p = x2 * tile_points;
  sint4 y2p = y2 * tile_points;

  // compute sectors

  sint4 sx1 = fine_grid(xp)/sec_wh;
  sint4 sy1 = fine_grid(yp)/sec_wh;
  sint4 sx2 = fine_grid(x2p)/sec_wh;
  sint4 sy2 = fine_grid(y2p)/sec_wh;

  // take the rectangle of motion sectors and mark them
  // a line drawing algorithm would be better

  if (sx2 < sx1) swap(sx1,sx2);
  if (sy2 < sy1) swap(sy1,sy2);

  if (sx1 < 0) sx1 = 0; else if (sx2 >= sec_nx) sx2 = sec_nx-1;
  if (sy1 < 0) sy1 = 0; else if (sy2 >= sec_ny) sy2 = sec_ny-1;

  for (sint4 i = sx1; i <= sx2; ++i) {
    for (sint4 j = sy1; j <= sy