#ifndef ScriptParser_H
#define ScriptParser_H

/** @file ScriptParser.H
    Reads a script and converts it into an executable ActionScript.

    $Id: ScriptParser.H,v 1.3 2006/04/03 01:53:42 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/ScriptParser.H,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "Script.H"
#include "SType.H"
#include "Game.H"
#include "Action.H"
#include "Token.H"

class ScriptObj;
class SimpleActions;
class PlayerInfo;
class Game;
class ObjActionDefinition;

//=============================================================================

class ScriptParser {

protected:
  std::istream *istrm;

  Tokenizer tokenizer;

  VarList *vlist; ///< Pointer to the ActionScript's data.

  /// @name Maps symbol names from a raw script to the vlist arrays
  //@{
  Vector<std::string> param_symbol;
  Vector<std::string> local_symbol;
  Vector<std::string> obj_symbol;
  Vector<std::string> global_symbol;
  //@}

  void check_redeclaration(const std::string &var);

  /// Pseudo program counters for stmtlists' eval functions.
  /// Counters are added and removed as the parser goes through the action.
  /// This list stores the PCs the current statement is operating under.
  std::list<uint4*> PCs;

  PlayerInfo *pi;
  Game *game;
  std::string file_name;
  sint4 line_num; // line in the file where the action is defined

  sint4 precedenceOf(const std::string &op) const;

public:

  static ScriptFunctionList functions;

  void parseError(const std::string &s = "") {
    errstr << "PARSE ERROR: " << current_action_name <<
      " : LINE " << tokenizer.cur_line() << " " << s << std::endl;
    ERR("");
  }

  std::string getToken();
  void putbackToken(const std::string &t);
  ScriptNode* getTerm(sint4 &constant);
  ScriptNode* parseTillLowOp(sint4 prec, sint4 *res_is_const, ScriptNode *lnode = 0, sint4 l_is_const = 0);
  ScriptNode* parseExp(const std::string &expectedEndMarker, ScriptNode *lnode = 0);
  ScriptNode* parseExp(const std::set<std::string> &EndMarkers, std::string &end_marker);

  VariableNode* getVariable(bool as_obj = false);
  ScriptNode* getAction(bool parse_delay);

  IfStmt*     parseIfStmt();
  WhileStmt*  parseWhileStmt();
  ForLoop*    parseForLoop();

  ScriptNode* parseReturnStmt();
  ScriptNode* parseBreakStmt();

  ScriptNode* parseObjLoop();
  ScriptNode* parseIntersectLoop();

  ScriptNode* parseStmt(const std::string *end_marker = 0);
  StmtList*   parseStmtList();

  VariableNode* parseObjAddr(bool has_attr, std::string *attr = 0);

  std::string current_action_name;

  ScriptParser() { istrm = 0; vlist = 0; pi = 0; game = 0; }

  void set_context(PlayerInfo &inf, Game &_game, std::string file = "", sint4 line = 0) {
    pi = &inf;
    game = &_game;
    file_name = file;
    line_num = line;
  }

  // read in a script and bind variables to addresses in a new ActionScript
  ActionScript* Parse(const ObjActionDefinition &def);

  ScriptNode* eval_expression(const std::string &exp);

  //-------------------------------------------------------------------
  /// @name Exposes new functions to the scripting language.
  //-------------------------------------------------------------------
  //@{
  template<class T>
  static void add(const ScriptFunctionDef &def, stype(T::*f)(const SActParam&) ) {
    functions.add_function(def, f);
  }
  template<class T>
  static void add(const std::string &name, sint4 num_obj_ids, sint4 num_params,
           stype(T::*f)(const SActParam&) ) {
    ScriptFunctionDef def;
    def.name = name;
    def.num_obj_ids = num_obj_ids;
    def.num_params = num_params;
    functions.add_function(def, f);
  }
  static void add(const ScriptFunctionDef &def, stype(*f)(const SActParam&) ) {
    functions.add_function(def, f);
  }
  static void add(const std::string &name, sint4 num_obj_ids, sint4 num_params,
           stype(*f)(const SActParam&) ) {
    ScriptFunctionDef def;
    def.name = name;
    def.num_obj_ids = num_obj_ids;
    def.num_params = num_params;
    functions.add_function(def, f);
  }
  //@}
};

//=============================================================================

#endif
