
# ifndef SR_MAT_H
# define SR_MAT_H

/** \file sr_mat.h 
 * A 4x4 matrix */

# include <SR/sr_vec.h>

/*! \class SrMat sr_mat.h
    \brief A 4x4 matrix

    The transformation methods consider that the matrix will be 
    applied to a line vector at the left side of the matrix. So 
    that in all matrix transformations M should multiply v as vM. 
    When using the matrix as a geometric transformation, the
    first three elements of the fourth line represent the 
    translation vector. This is to follow the OpenGL design. */
class SrMat
 { private :
    float e[16];
   public :
    static const SrMat null; //!< A null matrix
    static const SrMat id;   //!< An identity matrix
    enum InitializeMode { NotInitialized };
   public :

    /*! Default constructor initializes SrMat as identity. */
    SrMat () { set(id.e); }

    /*! Copy constructor. */
    SrMat ( const SrMat &m ) { set(m.e); }

    /*! Constructor without initialization. SrMat values are simply not
        initialized when declaring SrMat m(SrMat::NotInitialized) */
    SrMat ( InitializeMode /*m*/ ) {}

    /*! Constructor from a float[16] pointer. */
    SrMat ( const float *p );

    /*! Constructor from a double[16] pointer. */
    SrMat ( const double *p );

    /*! Constructor from 16 values. */
    SrMat ( float a, float b, float c, float d, float e, float f, float g, float h, 
            float i, float j, float k, float l, float m, float n, float o, float p );

    /*! Copy operator. */
    void operator= ( const SrMat& m ) { set(m.e); }

    /*! float pointer type cast operator */
    operator float*() { return &e[0]; }

    /*! const float pointer type cast operator */
    operator const float*() const { return &e[0]; }

    /*! Returns a float pointer to the SrMat element with given index. */
    float* pt ( int i ) { return &e[i]; }

    /*! Returns a const float pointer to the SrMat element with given index. */
    const float* pt_const ( int i ) const { return &e[i]; }

    /*! Permits access to an element of the matrix as a vector of 16 elements.
        The given index must be from 0 to 15, no chekings are done.
        Implemented inline. */
    float& operator[] ( int i ) { return e[i]; }
    
    /*! const version of [] operator */
    float operator[] ( int i ) const { return e[i]; }

    /*! const version of operator[], implemented inline */
    float get ( int i ) const { return e[i]; }

    /*! set element i in [0,16], implemented inline */
    void set ( int i, float v ) { e[i]=v; }

    /*! Permits access to an element of the matrix using a line and column
        indices. The given indices can be from 0 to 3. No indices chekings
        are done. Implemented inline. */
    float& operator() ( int i, int j ) { return e[i*4+j]; }

    /*! const version of operator(), implemented inline */
    float get ( int i, int j ) const { return e[i*4+j]; }

    float& e11 () { return e[0]; } //!< Returns a reference to the element [0]
    float& e12 () { return e[1]; } //!< Returns a reference to the element [1]
    float& e13 () { return e[2]; } //!< Returns a reference to the element [2]
    float& e14 () { return e[3]; } //!< Returns a reference to the element [3]
    float& e21 () { return e[4]; } //!< Returns a reference to the element [4]
    float& e22 () { return e[5]; } //!< Returns a reference to the element [5]
    float& e23 () { return e[6]; } //!< Returns a reference to the element [6]
    float& e24 () { return e[7]; } //!< Returns a reference to the element [7]
    float& e31 () { return e[8]; } //!< Returns a reference to the element [8]
    float& e32 () { return e[9]; } //!< Returns a reference to the element [9]
    float& e33 () { return e[10]; } //!< Returns a reference to the element [10]
    float& e34 () { return e[11]; } //!< Returns a reference to the element [11]
    float& e41 () { return e[12]; } //!< Returns a reference to the element [12]
    float& e42 () { return e[13]; } //!< Returns a reference to the element [13]
    float& e43 () { return e[14]; } //!< Returns a reference to the element [14]
    float& e44 () { return e[15]; } //!< Returns a reference to the element [15]

    /*! Sets all elements of SrMat from the given float[16] pointer. */
    void set ( const float *p );

    /*! Sets all elements of SrMat from the given double[16] pointer. */
    void set ( const double *p );

    /*! Sets the four elements of line 1. Implemented inline. */
    void setl1 ( float x, float y, float z, float w ) { e[0]=x; e[1]=y; e[2]=z; e[3]=w; }
    
    /*! Sets the four elements of line 2. Implemented inline. */
    void setl2 ( float x, float y, float z, float w ) { e[4]=x; e[5]=y; e[6]=z; e[7]=w; }
    
    /*! Sets the four elements of line 3. Implemented inline. */
    void setl3 ( float x, float y, float z, float w ) { e[8]=x; e[9]=y; e[10]=z; e[11]=w; }
    
    /*! Sets the four elements of line 4. Implemented inline. */
    void setl4 ( float x, float y, float z, float w ) { e[12]=x; e[13]=y; e[14]=z; e[15]=w; }

    /*! Sets the first three elements of line 1. Implemented inline. */
    void setl1 ( const SrVec &v ) { e[0]=v.x; e[1]=v.y; e[2]=v.z; }
    
    /*! Sets the first three elements of line 2. Implemented inline. */
    void setl2 ( const SrVec &v ) { e[4]=v.x; e[5]=v.y; e[6]=v.z; }
    
    /*! Sets the first three elements of line 3. Implemented inline. */
    void setl3 ( const SrVec &v ) { e[8]=v.x; e[9]=v.y; e[10]=v.z; }
    
    /*! Sets the first three elements of line 4. Implemented inline. */
    void setl4 ( const SrVec &v ) { e[12]=v.x; e[13]=v.y; e[14]=v.z; }

    /*! Returns true if all elements are equal to 0.0, false otherwise. */
    bool isnull () const { return *this==null; }

    /*! Returns true if the matrix is identical to SrMat::id, false otherwise. */
    bool isid () const { return *this==id;   }

    /*! Makes SrMatrix be a null matrix. */
    void zero () { *this=null; }

    /*! Make elements in interval [-epsilon,epsilon] to become 0 */
    void round ( float epsilon );

    /*! Makes SrMatrix be an identity matrix. */
    void identity () { *this=id; }

    /*! Transpose SrMatrix. */
    void transpose ();

    /*! Transpose the 3x3 sub matrix. */
    void transpose3x3 ();

    /*! Makes SrMat be a translation transformation, that can be applied to a vector x as xM. */
    void translation ( float tx, float ty, float tz );

    /*! Makes SrMat be a translation transformation, that can be applied to a vector x as xM. */
    void translation ( const SrVec &v ) { translation(v.x,v.y,v.z); }

    /*! Pre-multiplies SrMat with a translation matrix constructed with the 
        vector v. The multiplication is optimized considering that SrMat has
        values only on its 3x3 sub matrix. */
    void left_combine_translation ( const SrVec &v );

    /*! Pos-multiplies SrMat with a translation matrix constructed with the
        vector v. Here the first three values of the 4th line are added by v. */
    void right_combine_translation ( const SrVec &v );

    /*! Combines in SrMat a scale factor of sx, sy and sz on each coordinate. */
    void combine_scale ( float sx, float sy, float sz );

    /*! Combines in SrMat a global scale factor of s. */
    void combine_scale ( float s ) { combine_scale(s,s,s); }

    /*! Makes SrMat be a scale transformation, that can be applied to a vector x as xM. */
    void scale ( float sx, float sy, float sz );

    /*! Makes SrMat be a scale transformation, that can be applied to a vector x as xM. */
    void scale ( float s ) { scale(s,s,s); }

    /*! Makes SrMat be a rotation transformation around x axis. The given parameters 
        are the sine and cosine of the desired angle. The transformation can then be 
        applied to a vector x as xM. */
    void rotx ( float sa, float ca );

    /*! Makes SrMat be a rotation transformation around y axis. The given parameters 
        are the sine and cosine of the desired angle. The transformation can then be 
        