// $Id: PRATerrain.C,v 1.1 2006/12/31 04:09:40 orts_furtak Exp $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

#include "PRATerrain.H"
#include "PRAEngine.H"
#include "Global.H"
#include "Options.H"
#include "GameObj.H"
#include "ForceFieldBS.H"
#include <math.h>

using namespace std;

#define TEST(x)
//#define TEST(x) x

//-----------------------------------------------------------------------------
void PR_Terrain::add_options()
{
  ::Options o("SimpleTerrain");

  o.put("-gran",  sint4(2),  "granularity of the fine grid default 2x2");
  //o.put("-costonly", "represent boundaries only by higher edge cost");
   
}

//-----------------------------------------------------------------------------
void PR_Terrain::init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 me, sint4 neutral)
{
  TerrainBasicImp<PR_Task>::init(tiles_x_, tiles_y_, tile_points_, me, neutral);
  sint4 gran;
  Options::get("-gran", gran);
  pfEngine = boost::shared_ptr<PRAEngine>(new PRAEngine(tiles_x_, tiles_y_, tile_points_, gran));
}

//-----------------------------------------------------------------------------
void PR_Terrain::add_obj(const Object *obj)
{
  TerrainBasicImp<PR_Task>::add_obj(obj);
  assert(pfEngine!=NULL);
  pfEngine->insert_object(obj);
  objects[obj] = false;
}

//-----------------------------------------------------------------------------
void PR_Terrain::update_obj(const Object *obj)
{
  TerrainBasicImp<PR_Task>::update_obj(obj);
  assert(pfEngine!=NULL);
  pfEngine->update_object(obj);
}

//-----------------------------------------------------------------------------
void PR_Terrain::remove_obj(const Object *obj)
{
  TerrainBasicImp<PR_Task>::remove_obj(obj);
  assert(pfEngine!=NULL);
  pfEngine->remove_object(obj);
  objsToStop.erase(obj);
  objects.erase(obj);
}

//-----------------------------------------------------------------------------
void PR_Terrain::add_segments(const Vector<Segment> &segs)
{
  TerrainBasicImp<PR_Task>::add_segments(segs);
  assert(pfEngine!=NULL);
  pfEngine->insert_boundaries(segs);
  invalidate_paths(segs);
  FORALL(segs, i){
    segments.push_back(*i);
  }
}

//----------------------------------------------------------------------------------
void PR_Terrain::cancel_task(const Object *obj)
{
  TerrainBasicImp<PR_Task>::cancel_task(obj);
  objsToStop.insert(obj);
}

//----------------------------------------------------------------------------------
void PR_Terrain::remove_task(TaskId tid)
{
  FIND (taskId2task, it, tid);
  if( it != taskId2task.end() ) {
    Task& task = it->second->task;
    objsToStop.insert(task.objs.begin(), task.objs.end());
  }
  TerrainBasicImp<PR_Task>::remove_task(tid);
}


//----------------------------------------------------------------------------------
real8 PR_Terrain::find_path(const Loc &/*l1*/, const Loc &/*l2*/, sint4 /*radius*/, Path &/*path*/)
{
  assert(0); // unimplemented yet    
  return -1;
}

//----------------------------------------------------------------------------------
// fixme: we only check the path of the *center* of the unit vs the new boundaries,
// and we should account for its size (radius).
void PR_Terrain::invalidate_paths(const Vector<Segment>& /*segments*/)
{
 //  FORALL(tasks, ti) {
//     if( !ti->isValid() ) continue;
//     // fixme: better handling of zcat != ON_LAND ?
//     if( ti->path->obj->get_zcat() != Object::ON_LAND ) continue;

//     FORALL(segments, si) {
//       if( ti->path->is_intersect(*si) ) {
// 	// this path is invalid, mark as pending for (re)planning
// 	pendingTasks.push_back(ti);
// 	ti->isPending = true;
// 	cout << "Marked an invalid path" << endl;
// 	break;
//       }
//     }
//   }
}

//----------------------------------------------------------------------------------
void PR_Terrain::execute_tasks(Vector<TerrainBase::MoveCmd> &cmds, Vector<TerrainBase::StatusMsg> &msgs)
{
  //clear the reserved locations in the object map
  FORALL(objects, it){
    it->second = Loc(-1, -1);
  }
  // Execute all active tasks
  for (Tasks::iterator j = tasks.begin(); j != tasks.end();) 
    {
      Tasks::iterator i = j++;

      // Report failure of planning (no path at all)
      if( i->noPathFound ) {
        TerrainBase::StatusMsg s;
        // fixme: this is not robust to future changes in the meaning of TaskId.
        s.task_id = (TerrainBase::TaskId)&(*i);
        s.obj = *(i->task.objs.begin());
        s.type = TerrainBase::StatusMsg::NO_PATH_FAILURE;
        msgs.push_back(s);
        continue;
      }

      // Skip invalid tasks.
      if( !i->isValid() ) { 
        TEST(cout << *i->task.objs.begin() << " task not valid" << endl;)
	  continue;
      }

      const Object *obj = i->obj;

      // Skip tasks that are waiting for a retry
//       if (cur_path.sleeping()) {
//         TEST(cout << obj << " path sleeping" << endl;)
// 	  continue;
//       }

      // Skip objects for which we gave a (move) action that is still pending
      if (obj->is_pending_action()) {
        TEST(cout << obj << " pending action" << endl;)
	  continue;
     }

      bool atGoal = is_at_goal(obj, i->task);

      // Look ahead: if the object will get to a waypoint in the next tick,
      // give the next action now the prevent it from wasting a tick on stopping.
      // This will cut a single tick at each waypoint, though it will really prevent 
      // the object from stopping only if there's no (network or processing) lag
      // between this client and the server.
      // atWaypoint can be true only if the current waypoint is an intermediate and
      // not the *final* waypoint/goal.
      bool atWaypoint = i->current_target > 0 &&  // reverse order, goal==0
	is_at_location(obj, i->path.at(i->current_target));
      
      // Skip a moving object which is currently not at a waypoint/goal
      if( (obj->get_speed() > 0 || obj->get_moving() != 0) && !atWaypoint && !atGoal ) {
	
	Loc vec;
	bool deflected = calculate_vector(obj, i->path.at(i->current_target), vec);
	bool nearWaypoint = is_near_waypoint(obj, i->path.at(i->current_target));
	uint4 counter = 1;
	while (nearWaypoint && deflected && i->current_target >0){
	  Loc next_wp_vector = i->path.at(i->current_target - 1);
	  Loc &curr_wp = i->path.at(i->current_target);
	  real8 d = next_wp_vector.distance(curr_wp);
	  
	  next_wp_vector.x = next_wp_vector.x - curr_wp.x;
	  next_wp_vector.y = next_wp_vector.y - curr_wp.y;
	  
	  real8 k = ((real8)(counter * obj->get_max_speed())) / d;
	  
	  if (k > 1){
	    cout << "waypoint failure\n";
	    //fixme: dunno what to do here.
	    break;
	  }
	  next_wp_vector.x = round(k * next_wp_vector.x);
	  next_wp_vector.y = round(k * next_wp_vector.y);
	  
	  curr_wp.x = curr_wp.x + next_wp_vector.x;
	  curr_wp.y = curr_wp.y + next_wp_vector.y;
	  
	  deflected = calculate_vector(obj, i->path.at(i->current_target), vec);
	  nearWaypoint = is_near_waypoint(obj, i->path.at(i->current_target));
	  counter++;
	}

	if (vec.x != i->motion_vector.x || vec.y != i->motion_vector.y){
	  i->motion_vector = vec;
	  add_move_command(obj, i->motion_vector, cmds);
	}
	
	
	continue;
      }

      // ok, so it needs attention:
      // 1. maybe it had reached the goal
      // 2. maybe it's at a waypoint (with the lookahead)
      // 3. maybe it needs the first orders
      // 4. maybe it unexpectedly collided with something

      if( atGoal ) {
	
        TEST(cout << obj << " arrived" << endl;)
	  
	  TerrainBase::StatusMsg s;
        // fixme: this is not robust to future changes in the meaning of TaskId.
        s.task_id = (TerrainBase::TaskId)&(*i);
        s.obj = obj; 
        s.type = TerrainBase::StatusMsg::ARRIVED;
        msgs.push_back(s);

      }
      // if we are at a waypoint, move to next waypoint
      if( atWaypoint ){
	i->current_target--;
	//cur_path.no_orders_yet = false;
	
	Loc vec;
	calculate_vector(obj, i->path.at(i->current_target), vec);
	i->motion_vector = vec;
	add_move_command(obj, i->motion_vector, cmds);
	
      } 
      else {
	
	// Unexpected stop. The current policy is:
	// 1-5. Send the move command again.
	// 6. Third stop at the same place => report failure.
	
	Loc cur_position; obj->get_center(cur_position.x, cur_position.y);
	i->collision_step++;
	
       
	Loc vec;
	calculate_vector(obj, i->path.at(i->current_target), vec);
	i->motion_vector = vec;
	add_move_command(obj, i->motion_vector, cmds);
	
	if( i->collision_step > 5 ) {
	  TerrainBase::StatusMsg s;
	  // fixme: this is not robust to future changes in the meaning of TaskId.
	  s.task_id = (TerrainBase::TaskId)&(*i);
	  s.obj = obj;
	  s.type = TerrainBase::StatusMsg::MOVEMENT_FAILURE;
	  msgs.push_back(s);
        }
      }
    }
    
  // Stop objects (whose task was cancelled etc), and that did not yet 
  // get a new move command
  FORALL(cmds, c) {
    objsToStop.erase(c->obj);
  }
  FORALL(objsToStop, oi) {

    const Object *obj = *oi;
    /* // Not good, since our view might not be fully updated: --David
       if (obj->get_moving() == 0) {
       continue;
       }*/

    TerrainBase::MoveCmd m;
    m.obj = obj;
    obj->get_center(m.next_loc.x, m.next_loc.y); // doesn't really matter
    m.speed = 0; // 0 -> stop
    cmds.push_back(m);
  }
  objsToStop.clear();

  // remove finished tasks
  FORALL (msgs, t) {
    remove_task(t->task_id);
  }
}


void PR_Terrain::add_move_command(const Object *obj, const Loc &target, Vector<TerrainBase::MoveCmd> &cmds)
{ 
  TerrainBase::MoveCmd m;
  m.obj = obj;
  m.next_loc = target;
  m.speed = obj->get_max_speed();
  cmds.push_back(m);
  
  FIND (objects, it, obj){
    assert(it != objects.end() && "hrmm... all objects should be in objects map");
    Loc curr_loc;
    obj->get_center(curr_loc.x, curr_loc.y);
    Vec_2D vec(target.x - curr_loc.x, target.y - curr_loc.y);
    
    real8 k = ((real8)obj->get_max_speed()) / vec.length();
    vec.x = round(k * vec.x);
    vec.y = round(k * vec.y);
    
    it->second.x = curr_loc.x + vec.x;
    it->second.y = curr_loc.y + vec.y;
  }
}


//----------------------------------------------------------------------------------
bool PR_Terrain::plan_tasks(void)
{
  // plan a failed path
  bool did_something = plan_failed_task();

  // plan a pending (new) task
  if( !did_something )
    did_something = plan_pending_task();

  return did_something;
}

//----------------------------------------------------------------------------------

bool PR_Terrain::plan_pending_task(void)
{
  if (pendingTasks.empty()) return false;

  TIter tit = pendingTasks.front(); pendingTasks.pop_front();
  // mark the task as handled
  tit->isPending = false;

  // object to move
  assert(tit->task.objs.size() > 0);
  const Object *obj = *(tit->task.objs.begin());

  // if the object is moving, stop it first and come back to it later
  if (obj->get_speed() > 0 || obj->get_moving()) {
    objsToStop.insert(obj);
    pendingTasks.push_back(tit);
    return false;
    // fixme: might be better to try another pending task already in this tick...
  }

  // plan
  Path newpath;
  bool found;
  if (tit->task.target == Task::LOCATION) {
    found = pfEngine->find_path(obj, tit->task.goal_loc, newpath);
  }
  else {
    assert (tit->task.target == Task::OBJ);
    found = pfEngine->find_path(obj, tit->task.goal_obj, newpath);
  }

  // insert results into data structures
  if( found ) {
    assert(newpath.locs.size() > 0);
    tit->init_path(obj, newpath.locs);
  }
  else {
    // fixme: maybe add to "failed"/sleep etc. ?
    tit->noPathFound = true;
  }
  return true;
}


<<<<<<< PRATerrain.C
bool PR_Terrain::plan_failed_task(void)
{
  assert(pfEngine!=NULL);
=======
  //----------------------------------------------------------------------------------
  bool PR_Terrain::plan_tasks(sint4 max_time)
  {
    // plan a failed path
    bool did_something = plan_failed_task();
>>>>>>> 1.5

  // find a failed path for which we don't yet know that 'noPathFound'
  // (ie we still didn't try to replan and found that there is no path)
  FORALL_B(tasks,tit) {
    if( tit->isFailed && !tit->noPathFound )
      break;
  }
  // if none exists
  if( tit == tasks.end() )
    return false;

  // object to move
  assert(tit->task.objs.size() > 0);
  const Object *obj = *(tit->task.objs.begin());

  // if the object is moving, stop it first and come back to it later
  if( obj->get_speed() > 0 || obj->get_moving() ) {
    objsToStop.insert(obj);
    return false;
    // fixme: might be better to try another failed task already in this tick...
  }

  Path newpath;
  bool found;
  if (tit->task.target == Task::LOCATION) {
    found = pfEngine->find_path(obj, tit->task.goal_loc, newpath);
  }
  else {
    assert (tit->task.target == Task::OBJ);
    found = pfEngine->find_path(obj, tit->task.goal_obj, newpath);
  }

  // insert results into data structures
  if( found ) {
    assert(newpath.locs.size() > 0);
    tit->init_path(obj, newpath.locs);
  }
  else {
    // fixme: maybe add to "failed"/sleep etc. ?
    tit->noPathFound = true;
  }
  return true;
    
}


bool PR_Terrain::is_at_location(const Object* obj, const Loc& target)
{
  Loc position; obj->get_center(position.x, position.y);
  return position.distance(target) <= obj->get_speed();
}

bool PR_Terrain::is_near_waypoint(const Object* obj, const Loc& target){
  Loc position; obj->get_center(position.x, position.y);
  return position.distance(target) <= obj->get_speed()*WAYPOINT_PROX;
}

bool PR_Terrain::is_at_goal(const Object *obj, const TerrainBase::Task &task)
{
  // Getting to the goal depends on the definition of the task:
  if( task.target == Task::OBJ ) {
    switch(task.mode) {
    case Task::TOUCH :
      return( obj->distance_to(*task.goal_obj) <= 1 );
      // using 1-tick look-ahead:
      //return( obj->distance_to(*task.goal_obj) <= obj->get_speed() );
      break;

    case Task::VICINITY :
             
      assert(0 && "Not implemented yet");
      break;// $Id: PRATerrain.C,v 1.1 2006/12/31 04:09:40 orts_furtak Exp $

    case Task::ATTACK :

      return(is_in_range(obj, task.goal_obj));
             
      break;

    default:
      ERR("Unknown mode");
    }
  } else if( task.target == Task::LOCATION ) {
    switch(task.mode) {
    case Task::TOUCH :
      return is_at_location(obj, task.goal_loc);
      break;

    case Task::VICINITY :
	      
      assert(0 && "Not implemented yet");
      break;

    case Task::ATTACK :
      assert(0 && "Not implemented yet");
      break;

    default:
      ERR("Unknown mode");
    }
  } else {
    ERR("Unknown target type");
  }
  return false;
}

bool PR_Terrain::is_in_range(const Object *moving_obj, const Object *goal_obj){
  //fixme: no check for sight
  const GameObj *gob = dynamic_cast<const GameObj*>(moving_obj);
  sint4 range = 1;
  if (gob){
    ScriptObj *weapon = gob->component("weapon");
    const sintptr *r = weapon->get_const_int_ptr("max_ground_range");
    if (r){
      range = *r;
    }
  }

  //if not a GameObj or with no weapon range, should behave exactly
  //like the check for touch.
  return ( moving_obj->distance_to(*goal_obj) <= range );
    
  // this is the check in ShooterAI. Other way makes more intuitive sense though...
  //     real8 rdist = moving_obj->distance_to(*goal_obj);
  //     sint4 sdist = (sint4)floor(rdist);
  //     return (sdist <= range);
}


// void PR_Terrain::calculate_vector(const Object *obj, const TerrainBase::Loc &target, TerrainBase::Loc &output){
//   output = target;
//   Vector<const Object*> obstacles;
//   find_obstacles(obj, obstacles);
  
//   if (obstacles.size() <= 0){
//     return;
//   }
  
//   assert(obj->get_shape() == Object::CIRCLE && "Only Circle Movement implemented");
//   TerrainBase::Loc center;
//   obj->get_center(center.x, center.y);
  
//   real8 d_targ = center.distance(target);
  
//   FORALL(obstacles, i){
//     TerrainBase::Loc vec;
//     if ((*i)->get_shape() == Object::CIRCLE){
//       (*i)->get_center(vec.x, vec.y);
//       real8 d_obj = center.distance(vec);
      
//       vec.y = center.y - vec.y;
//       vec.x = center.x - vec.x;
       
//       real8 k = d_targ/d_obj;
      
//       vec.x = (sint4)(k * vec.x);
//       vec.y = (sint4)(k * vec.y);
      
//       output.x = output.x + vec.x;
//       output.y = output.y + vec.y;
//     }
//     else if ((*i)->get_shape() == Object::RECTANGLE){
//       vec = get_rect_repulsion(obj, *i); 
//     }
    
//   }
// }

bool PR_Terrain::calculate_vector(const Object *obj, const TerrainBase::Loc &target, TerrainBase::Loc &output){
  //output = target;
  
  assert(obj->get_shape() == Object::CIRCLE && "Only Circle Movement implemented");
  TerrainBase::Loc center, undeflected;
  obj->get_center(center.x, center.y);
  
  real8 d_targ = center.distance(target);
  
  output.x = target.x - center.x;
  output.y = target.y - center.y;
  
  Vec_2D target_vec(output.x, output.y);

  real8 k = ATTRACTOR_FORCE / d_targ;

  output.x = round(k * output.x);
  output.y = round(k * output.y);

  undeflected = output;

  FORALL(objects, i){
    Loc vec;
    if (i->first->get_shape() == Object::CIRCLE){
      Loc obstacle_center;
      if (i->second == Loc(-1,-1)){
	i->first->get_center(obstacle_center.x, obstacle_center.y);
      }
      else{
	obstacle_center = i->second;
      }
      vec = get_point_repulsion(obj, obstacle_center, i->first->get_radius());
      if (vec.x != 0 || vec.y != 0){
	output.x = output.x + vec.x;
	output.y = output.y + vec.y;
      }
      
    }
    else if (i->first->get_shape() == Object::RECTANGLE){
      Loc u_l, u_r, l_r, l_l; //upper-left, upper-right, lower-right, lower-left
      i->first->get_p1(u_l.x, u_l.y);
      i->first->get_p2(l_r.x, l_r.y);
      u_r.x = l_r.x;
      u_r.y = u_l.y;
      
      l_l.x = u_l.x;
      l_l.y = l_r.y;

      Segment seg;
      seg.l1 = u_l;
      seg.l2 = u_r;
      
      vec = get_line_repulsion(obj, seg);
      if (vec.x != 0 || vec.y != 0){
	//normalize_vector(vec, d_targ);
	output.x = output.x + vec.x;
	output.y = output.y + vec.y;
      }

      seg.l1 = u_r;
      seg.l2 = l_r;
      
      vec = get_line_repulsion(obj, seg);
      if (vec.x != 0 || vec.y != 0){
	//normalize_vector(vec, d_targ);
	output.x = output.x + vec.x;
	output.y = output.y + vec.y;
      }
      
      seg.l1 = l_r;
      seg.l2 = l_l;
      
      vec = get_line_repulsion(obj, seg);
      if (vec.x != 0 || vec.y != 0){
	//	normalize_vector(vec, d_targ);
	output.x = output.x + vec.x;
	output.y = output.y + vec.y;
      }

      seg.l1 = l_l;
      seg.l2 = u_l;
      
      vec = get_line_repulsion(obj, seg);
      if (vec.x != 0 || vec.y != 0){
	//normalize_vector(vec, d_targ);
	output.x = output.x + vec.x;
	output.y = output.y + vec.y;
      }
    }
  }
  
  FORALL(segments, seg_it){
    Loc vec;
    vec = get_line_repulsion(obj, *seg_it);
    if (vec.x != 0 || vec.y != 0){
      //normalize_vector(vec, d_targ);
      output.x = output.x + vec.x;
      output.y = output.y + vec.y;
    }
  }

  if (output == undeflected){
    output = target;
    return false;
  }
  
  Vec_2D direction_vec(output.x, output.y);
  
  restrict_vector(target_vec, direction_vec);
  
  //if the direction vector is 0, set it go at a right angle to the target
  //(should be always right or always left, but I can't figure out which).
  if (direction_vec.x == 0 && direction_vec.y == 0){
    direction_vec.x = target_vec.y;
    direction_vec.y = -target_vec.x;
  }
  
  output.x = center.x + direction_vec.x;
  output.y = center.y + direction_vec.y;

  return true;

}

void PR_Terrain::restrict_vector(const Vec_2D &target_vec, Vec_2D &vec){
  
  real8 dot_prod = vec.dot_prod(target_vec);
  if (dot_prod < 0){
    Vec_2D perpendicular(target_vec.y, -target_vec.x);
    real8 k = perpendicular.dot_prod(vec) / perpendicular.dot_prod(perpendicular);
    vec.x = round(k * perpendicular.x);
    vec.y = round(k * perpendicular.y);
  }
  
}


void PR_Terrain::normalize_vector(Loc &vec, const real8 &length){
  real8 d_vec = sqrt(vec.x * vec.x + vec.y * vec.y);
  real8 k = length / d_vec;
  vec.x = round(k * vec.x);
  vec.y = round(k * vec.y);
}


TerrainBase::Loc PR_Terrain::get_point_repulsion (const Object *moving_obj, const Loc &loc, sint4 radius){
  TerrainBase::Loc vec, center;
  moving_obj->get_center(center.x, center.y);
  real8 d = center.distance(loc) - moving_obj->get_radius() - radius +1;

  if (d < moving_obj->get_max_speed() * FIELD_RANGE_MULTIPLIER){
    vec.x = center.x - loc.x;
    vec.y = center.y - loc.y;
    real8 k = REPULSOR_BASELINE / (d*d*d);
    vec.x = round(k * vec.x);
    vec.y = round(k * vec.y);
  }
  else {
    vec.x = 0;
    vec.y = 0;
  }
  return vec;
}

TerrainBase::Loc PR_Terrain::get_line_repulsion (const Object *moving_obj, const Segment &seg){
  // q - (((q - p) dot n) / (n dot n))n

  Loc q, p, n, point_on_line;
  sint4 qp_dot_n, n_dot_n;
  real8 r;
  moving_obj->get_center(q.x, q.y);
  p = seg.l1;
  n.y = -(seg.l1.x - seg.l2.x);
  n.x = seg.l1.y - seg.l2.y;
  
  //q-p
  p.x = q.x - p.x;
  p.y = q.y - p.y;
  
  //(q-p) dot n
  qp_dot_n = p.x* n.x + p.y * n.y;
  
  //n dot n
  n_dot_n = n.x*n.x + n.y*n.y;
  
  r = ((real8)(qp_dot_n)) / ((real8)n_dot_n);
  
  n.x = round(r * n.x);
  n.y = round(r * n.y);

  point_on_line.x = q.x - n.x;
  point_on_line.y = q.y - n.y;
  
  Loc vec_to_l1, vec_to_l2;
  vec_to_l1.x = seg.l1.x - point_on_line.x;
  vec_to_l1.y = seg.l1.y - point_on_line.y;
  
  vec_to_l2.x = seg.l2.x - point_on_line.x;
  vec_to_l2.y = seg.l2.y - point_on_line.y;
  
  if ((vec_to_l1.x * vec_to_l2.x)>=0 && (vec_to_l1.y * vec_to_l2.y)>=0){
    real8 d_l1 = q.distance(seg.l1) - moving_obj->get_radius() +1;
    real8 d_l2 = q.distance(seg.l2) - moving_obj->get_radius() +1;
    if (d_l1 < d_l2){
      if (d_l1 < moving_obj->get_max_speed()*FIELD_RANGE_MULTIPLIER)
	return get_point_repulsion(moving_obj, seg.l1);
    }
    else if (d_l2 < moving_obj->get_max_speed()*FIELD_RANGE_MULTIPLIER)
      return get_point_repulsion(moving_obj, seg.l2);
  }
  else{
    real8 d = q.distance(point_on_line) - moving_obj->get_radius() +1;
    if (d < moving_obj->get_max_speed()*FIELD_RANGE_MULTIPLIER){
      real8 k = REPULSOR_BASELINE / (d*d*d);
      n.x = round(k * n.x);
      n.y = round(k * n.x);
      return n;
    }
  }
  Loc zero(0, 0);
  return zero;
}

sint4 PR_Terrain::round(real8 real){
  real8 i;
  i = floor(real);
  if (real-i >= 0.5){
    i = i+1;
  }
  return (sint4)i;
}

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

#define EPS 1e-9f


// UnitPath::UnitPath(const Object *obj_, const Vector<TerrainBase::Loc> &path_)
//   : obj(obj_), path(path_), no_orders_yet(true), collision_step(0), wakeup(0)
// {
//   current_target = path.size()-1;
//   if (path.size() > 0) {
//     goal = path[0];
//   }
//   spot = TerrainBase::Loc(-1, -1);
// }

// bool UnitPath::is_intersect(const TerrainBase::Segment &segment) const
// {
//   typeof(path.begin()) pj = path.begin();
//   typeof(path.begin()) pi = pj++;
//   for( ; pj != path.end() ; pi++, pj++ ) {
//     if( TerrainBase::Segment(*pi, *pj).touches(segment) )
//       return true;
//   }
//   return false;
// }

// #if 0
// Vector<Segment> UnitPath::intersect(const TerrainBase::Segment &segment)
// {
//   Vector<Segment> lines;
//   const Loc& a = segment.l1;
//   const Loc& b = segment.l2;
  
//   for (uint4 i = 1; i < path.size(); i++) {

//     Loc c = path[i-1];
//     Loc d = path[i];
    
//     real8 m1 = (real8)(c.y-d.y) / (real8)(c.x-d.x + EPS);
//     real8 m2 = (real8)(a.y-b.y) / (real8)(a.x-b.x + EPS);
    
//     if ((int)m1 == (int)m2) continue; // fixme: equality for reals doesn't work
    
//     real8 b1 = c.y - m1 * c.x;
//     real8 b2 = a.y - m2 * a.x;        
    
//     real8 X = (b2-b1) / (m1-m2);
//     real8 Y = m1*X + b1;
    
//     real8 t1x = (real8)(X - a.x) / ((real8)(b.x - a.x + EPS));
//     real8 t1y = (real8)(Y - a.y) / ((real8)(b.y - a.y + EPS));
//     real8 t2x = (real8)(X - c.x) / ((real8)(d.x - c.x + EPS));
//     real8 t2y = (real8)(Y - c.y) / ((real8)(d.y - c.y + EPS));
    
//     bool overlap1 = (t1x < 0 || t1x > 1);
//     if (!overlap1) continue;
    
//     bool overlap2 = (t1y < 0 || t1y > 1);
//     if (!overlap2) continue;    
    
//     bool overlap3 = (t2x < 0 || t2x > 1);
//     if (!overlap3) continue;        

//     bool overlap4 = (t2y < 0 || t2y > 1);
//     if (!overlap4) continue;    

//     lines.push_back(Segment(path[i-1], path[i]));
//   }
  
//   return lines;
// }
// #endif

PR_Task::PR_Task(const TerrainBase::Task& task_)
  : task(task_), isPending(true), isFailed(false), noPathFound(false),
    current_target(-1),  collision_step(0), motion_vector(-1, -1) 
{
  obj = 0;
}

void PR_Task::init_path(const Object *obj_, const Vector<TerrainBase::Loc> &path_){
  obj = obj_;
  path = path_;
  current_target = path.size() - 1;
  isFailed = false;
  isPending = false;
  noPathFound = false;
}
//-----------------------------------------------------------------------------------

Vec_2D Vec_2D::operator+(const Vec_2D &other) const{
  return Vec_2D(x + other.x, y + other.y);
}

Vec_2D Vec_2D::operator-(const Vec_2D &other) const{
  return Vec_2D(x - other.x, y - other.y);
}

sint4 Vec_2D::dot_prod(const Vec_2D &other) const{
  return x*other.x + y*other.y;
}

real8 Vec_2D::length() const{
  return sqrt(x*x + y*y);
}
