// $Id: SimpleTerrain.C,v 1.3 2007/05/26 05:31:06 orts_mburo Exp $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

#include "SimpleTerrain.H"
#include "ST_PFEngine.H"
#include "Options.H"
#include "GameObj.H"
#include "Collision2D.H"
#include <utility>
#include <iterator>
#include <math.h>
#include "Profiler.H"
//#include "boost\smart_ptr.hpp"

using namespace std;
using boost::scoped_ptr;

#define TEST(x)
//#define TEST(x) x
#define TEST2(x)
//#define TEST2(x) x


namespace SimpleTerrain {

  const uint4 ST_Terrain::MAXIMUM_TASK_AGE = 250;          // milliseconds
  const sint4 ST_Terrain::ANTICIPATE_COLLISION_TICKS = 16; // 2 seconds * 8 ticks_per_second
  const sint4 ST_Terrain::FINE_GRID_RES_BITS = 6;
  const uint4 ST_Terrain::TILES_PER_MOTION_SECTOR = 3;

  //-----------------------------------------------------------------------------
  static Loc get_center_loc(const Object *obj)
  {
    Loc l;
    obj->get_center(l.x, l.y);
    return l;
  }

  static ScalarPoint get_center_sp(const Object *obj)
  {
    ScalarPoint l;
    obj->get_center(l.x, l.y);
    return l;
  }

  static real8 path_length(Path &path)
  {
    if( path.locs.size() < 2 )
      return 0;
    real8 total = 0;
    for( Vector<Loc>::iterator i = ++path.locs.begin() ; i != path.locs.end() ; i++ )
      total += i->distance(*(i-1));
    return total;
  }
    
  // Assuming path is in the usual reversed order, with goal at locs[0]
  static real8 path_length_from(Path &path, const Loc& start_loc)
  {
    if( path.locs.size() < 1 ) 
      return 0;
    return start_loc.distance(path.locs.back()) + path_length(path);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::add_options()
  {
    ::Options o("SimpleTerrain");
    o.put("-costonly", "represent boundaries only by higher edge cost");
    o.put("-highres", "use a high resolution world map (slower but more accurate)");
    o.put("-nomotionsectors", "don't use motion sectors for collision detection (test all pairs of units)");
  }

  //-----------------------------------------------------------------------------
  ST_Terrain::ST_Terrain() {
    sint4 seed = 0;
    ::Options::get("-seed", seed);
    rand.srand(seed);

    noMotionSectors = false;
    ::Options::get("-nomotionsectors", noMotionSectors);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_)
  {
    TerrainBasicImp<ST_Task>::init(tiles_x_, tiles_y_, tile_points_, client_pID_, neutral_pID_);
    
    // initialize the internal PFEngine
    Vector<sint4> levels_of_detail;
    bool high_resolution_only;
    Options::get("-highres", high_resolution_only);
    if( !high_resolution_only ) {
      levels_of_detail.push_back(4);  // default - use a resolution reduced by a factor of 4.
    }
    levels_of_detail.push_back(1);  // also (or exclusively) maintain a full resolution view.
    pfEngine = boost::shared_ptr<PFEngine>(new PFEngine(tiles_x_, tiles_y_, tile_points_, levels_of_detail));

    // initialize motion squares 
    //cout << "Created " << tiles_x_ * tiles_y_ << " sectors" << endl; 
/*    Vector<Sector> sectors_(tiles_x_ * tiles_y_);
    sectors = sectors_;
    sect_nx = tiles_x_;
  */  
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::add_obj(const Object *obj)
  {
    TerrainBasicImp<ST_Task>::add_obj(obj);
    assert(pfEngine!=NULL);
    pfEngine->insert_object(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::update_obj(const Object *obj)
  {
    TerrainBasicImp<ST_Task>::update_obj(obj);
    assert(pfEngine!=NULL);
    pfEngine->update_object(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::remove_obj(const Object *obj)
  {
    // shorthand
    typedef ObjIsGoalInTaskId::iterator OGIT;

    // 