//-----------------------------------------------------------------------------
/** @file bucketfringesearch.h
    FringeSearch using multiple buckets.

    $Id: bucketfringesearch.h,v 1.1 2006/12/31 04:09:41 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/pathfinding/ua_path_lib/src/bucketfringesearch.h,v $
*/
//-----------------------------------------------------------------------------

#ifndef PATHFIND_BUCKETFRINGESEARCH_H
#define PATHFIND_BUCKETFRINGESEARCH_H

#include <limits>
#include "marker.h"
#include "search.h"
#include "util.h"

//-----------------------------------------------------------------------------

namespace PathFind
{
    using namespace std;

    /** Statically allocated set of linked lists used in BucketFringeSearch. */
    class BucketFringe
    {
    public:
        BucketFringe(int numberBuckets);
        
        int getCurrentNode() const
        {
            return m_currentNode;
        }
        
        void init(int numberNodes);
        
        /** Insert node after current node. */
        void insertNode(int nodeId);
        
        /** Insert node at front in bucket. */
        void insertNode(int nodeId, int bucket);

        bool isEmpty(int bucket)
        {
            return (getHeadNode(bucket) == NO_NODE);
        }
        
        void nextNode();
        
        void print(ostream& o) const;
        
        void removeCurrentNode()
        {
            removeNode(m_currentNode);
        }
        
        void startIteration(int bucket);
        
    private:
        /** Node used in BucketFringe. */
        class Node
        {
        public:
            int m_next;
            
            int m_previous;
        };
        

        int m_currentBucket;

        int m_currentNode;
        
        int m_numberBuckets;

        int m_numberNodes;
        
        vector<int> m_headNode;
        
        /** Stores bucket number for each node in the environment.
            Contains -1 for nodes not in any bucket.
        */
        vector<int> m_nodeInBucket;
        
        vector<Node> m_nodes;
        
        
        bool allLowerBucketsEmpty(int bucket) const;

        int getHeadNode(int bucket) const
        {
            assert(bucket >= 0);
            assert(bucket < m_numberBuckets);
            return m_headNode[bucket];
        }

        Node& getNode(int nodeId)
        {
            assert(nodeId >= 0 && nodeId < m_numberNodes);
            return m_nodes[nodeId];
        }
        
        const Node& getNode(int nodeId) const
        {
            assert(nodeId >= 0 && nodeId < m_numberNodes);
            return m_nodes[nodeId];
        }
        
        void removeNode(int nodeId);

        void setHeadNode(int bucket, int nodeId)
        {
            assert(bucket >= 0);
            assert(bucket < m_numberBuckets);
            m_headNode[bucket] = nodeId;
        }
    };

    /** FringeSearch using multiple buckets.

        %Statistics (see Search::createStatistics and Search::getStatistics):
        - "cpu_time"
        - "iterations"
        - "nodes_expanded"
        - "nodes_visited"
        - "path_cost"
        - "path_length"
    */
    template<class MARKER = MarkerFastClear>
    class BucketFringeSearch
        : public Search
    {
    public:
        BucketFringeSearch(int numberBuckets, int deltaBucketFValue);

        StatisticsCollection createStatistics();

        bool findPath(const Environment& env, int start, int target);

        const vector<int>& getPath() const
        {
            return m_path;
        }

        const StatisticsCollection& getStatistics() const;

        const vector<char>& getVisitedNodes() const
        {
            return m_visitedNodes;
        }

    private:   
        /** Cache element for g-value used in BucketFringeSearch. */
        class CacheElement
        {
        public:
            int m_gValue;
            
            int m_iteration;

            int m_parent;
        };


        static const bool BUCKETFRINGE_DEBUG = false;
        
        int m_deltaBucket