//-----------------------------------------------------------------------------
/** @file idastar.C
    @see idastar.h

    $Id: idastar.C,v 1.1 2006/12/31 04:09:41 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/pathfinding/ua_path_lib/src/idastar.C,v $
*/
//-----------------------------------------------------------------------------

#include "idastar.h"

#include <cassert>
#include <iostream>
#include <math.h>
#include "util.h"

using namespace std;
using namespace PathFind;

//-----------------------------------------------------------------------------

IDAStar::IDAStar(Caching caching, bool nodeCountForAllTargets)
    : m_nodeCountForAllTargets(nodeCountForAllTargets),
      m_caching(caching),
      m_statistics(createStatistics()),
      m_branchingFactor(m_statistics.get("branching_factor"))
{
}

bool IDAStar::checkCache(int nodeId, int /*f*/, int g, int iteration, int& fMin)
{
    if (m_cache[nodeId].m_iter > 0) 
    {
        if ((m_cache[nodeId].m_g < g) ||
            ((m_cache[nodeId].m_g == g) && 
             (m_cache[nodeId].m_iter == iteration)))
        {
            fMin = numeric_limits<int>::max();
            return true;
        }
        if (m_caching == F_CACHING)
        {
            // Check if backed-up F-value causes an early termination.
            if (m_cache[nodeId].m_f > m_fLimit)
            {
                fMin = m_cache[nodeId].m_f;
                return true;
            }
        }

    }
    return false;
}

StatisticsCollection IDAStar::createStatistics()
{
    StatisticsCollection collection;
    collection.create("aborted");
    collection.create("cpu_time");
    collection.create("path_cost");
    collection.create("path_length");
    collection.create("branching_factor");
    collection.create("nodes_expanded");
    collection.create("nodes_visited");
    collection.create("nodes_expanded_last_iter");
    collection.create("nodes_visited_last_iter");
    collection.create("iterations");
    return collection;
}

bool IDAStar::findPath(const Environment& env, int start, int target)
{
    assert(env.isValidNodeId(start));
    assert(env.isValidNodeId(target));
    Clock clock;
    m_statistics.clear();
    m_nodesExpanded = 0;
    m_nodesVisited = 0;
    m_env = &env;
    m_target = target;
    m_path.clear();
    m_abortSearch = false;
    m_visitedNodes.resize(env.getNumberNodes());
    fill(m_visitedNodes.begin(), m_visitedNodes.end(), ' ');
    if (m_caching != NO_CACHING)
    {
        m_cache.clear();
        m_cache.resize(env.getNumberNodes());
    }
    findPathIdaStar(start);
    addStatistics("aborted", m_abortSearch ? 1 : 0);
    addStatistics("cpu_time", clock.get());
    addStatistics("nodes_expanded", m_nodesExpanded);
    addStatistics("nodes_visited", m_nodesVisited);
    addStatistics("nodes_expanded_last_iter", m_nodesExpandedCurrentIter);
    addStatistics("nodes_visited_last_iter", m_nodesVisitedCurrentIter);
    addStatistics("path_length", m_path.size());
    return true;
}

void IDAStar::findPathIdaStar(int start)
{
    const int maxFLimit = m_env->getNumberNodes() * m_env->getMaxCost();
    int heuristic = m_env->getHeuristic(start, m_target);
    m_fLimit = heuristic;
    int minCost = m_env->getMinCost();
    unsigned int expectedMaxDepth =
        static_cast<unsigned int>(heuristic / minCost);
    if (m_successorStack.size() < expectedMaxDepth)
        m_successorStack.resize(expectedMaxDepth);
    int iteration = 1;
    m_targetFound = false;
    while (true)
    {
        m_nodesExpandedCurrentIter = 0;
        m_nodesVisitedCurrentIter = 0;
        //cerr << "Iteration " << iteration << " limit = " << m_fLimit << '\n';
        int fMin = numeric_limits<int>::max();
        searchPathIdaStar(iteration, start, NO_NODE, 0, 0, fMin);
        m_nodesExpanded += m_nodesExpandedCurrentIter;
        m_nodesVisited += m_nodesVisitedCurrentIter;
        if (m_targetFound)
            break;
        if (fMin > maxFLimit)
            break;
        m_fLimit = fMin;
        ++iteration;
    }
    addStatistics("iterations", iteration);
}

const StatisticsCollection& IDAStar::getStatistics() const
{
    return m_statistics;
}

void IDAStar::markVisitedNode(int nodeId, int iteration)
{
    if (m_visitedNodes[nodeId] != ' ')
        return;
    m_visitedNodes[nodeId] = getVisitedNodeLabel(iteration);
}

bool IDAStar::searchPathIdaStar(int iteration, int node, int lastNode,
                                int depth, int g, int& fMin)
{
    ++m_nodesVisitedCurrentIter;
    markVisitedNode(node, iteration);
    if (m_nodesLimit >= 0 && m_nodesVisited > m_nodesLimit)
    {
        m_abortSearch = true;
        fMin = numeric_limits<int>::max();
        return false;
    }
    int f = g + m_env->getHeuristic(node, m_target);

    if (f > m_fLimit)
    {
        fMin = f;
        return false;
    }
    if (node == m_target)
    {
        if (! m_targetFound)
        {
            addStatistics("path_cost", f);
            fMin = f;
            m_targetFound = true;
        }
        if (! m_nodeCountForAllTargets)
        {
            m_path.push_back(node);
            return true;
        }
    }

    //cerr << "Visit " << node << " (" << f << ")\n";
    // Checking the cache before the fLimit is more efficient
    // and should make results (number of iterations, nodes in path)
    // equal to FringeSearch.
    // Checking the cache after the fLimit should make
    // results equal to IDAStar without caching.
    if ((m_caching != NO_CACHING) && (node != m_target))
    {
        if (checkCache(node, f, g, iteration, fMin))
            return false;
        addCache(node, f, g, iteration);
    }

    ++m_nodesExpandedCurrentIter;
    assert(depth >= 0);
    if (m_successorStack.size() < static_cast<unsigned int>(depth + 1))
        m_successorStack.resize(depth + 1);
    vector<Environment::Successor>& successors = m_successorStack[depth];
    m_env->getSuccessors(node, lastNode, successors);
    int numberSuccessors = successors.size();
    m_branchingFactor.add(numberSuccessors);
    fMin = numeric_limits<int>::max();
    for (int i = 0; i < numberSuccessors; ++i)
    {
        // Get reference on edge again, because resize could have changed it.
        const Environment::Successor& successor = m_successorStack[depth][i];
        int targetNodeId = successor.m_target;
        int fMinChild;
        if (targetNodeId == lastNode)
            continue;
        if (searchPathIdaStar(iteration, targetNodeId, node, depth + 1,
                              g + successor.m_cost, fMinChild))
        {
            if (fMinChild < fMin)
                fMin = fMinChild;
            m_path.push_back(node);
            return true;
        }
        if (fMinChild < fMin)
            fMin = fMinChild;

        if (m_abortSearch)
            return false;
    }
    if (m_caching == F_CACHING)
    {
        if (fMin > m_cache[node].m_f)
            m_cache[node].m_f = fMin;
    }    
    return false;
}

//-----------------------------------------------------------------------------
