// $Id: Server.C,v 1.35 2007/05/25 01:53:28 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "Server.H"
#include "Game.H"
#include "TCP_Client.H"
#include "Compress.H"
#include "CycleCounter.H"
#include "ShutDown.H"
#include "ByteData.H"
#include "Options.H"
#include "IO_Buffer.H"

using namespace std;

const sint4 Server::FROM        = EventFactory::new_who();
const sint4 Server::READY_MSG   = EventFactory::new_what();    // sent when game loop starts
const sint4 Server::STEP_MSG    = EventFactory::new_what();     // sent after views are sent

const sint4 AUTH_WAIT_SEC = 3;

Server::Options::Options()
{
  ::Options::get("-clev", clev);
  ::Options::get("-synchro", synchro);
  ::Options::get("-port", port);
  ::Options::get("-f", freq);
  ::Options::get("-wait", wait);
  ::Options::get("-sdelay", sdelay);
  ::Options::get("-tmax", tmax);
  ::Options::get("-act", act);
  ::Options::get("-keepalive", keepalive);
  ::Options::get("-seed", seed);
  ::Options::get("-loadreplay", load_replay);
  ::Options::get("-savereplay", save_replay);
  ::Options::get("-loadfile", load_file);
  ::Options::get("-savefile", save_file);
  ::Options::get("-ids", ids);
}

void Server::Options::add()
{
  ::Options o("Server");
  o.put("-clev",    sint4(5),     "compression level");
  o.put("-synchro",               "wait for client messages");

  o.put("-port",    sint4(7777),  "port to connect to");
  o.put("-f",       real8(8),     "simulation frequency");
  o.put("-wait",    sint4(0),     "<0:keyboard input, 0:full speed, >0:msec");
  o.put("-sdelay",  sint4(0),     "start delay in seconds");
  o.put("-tmax",    sint4(-1),    "max # of ticks (<0:inf)");

  o.put("-act",                   "generate actions");
  o.put("-keepalive",             "keep simulating after all clients disconnect");
  
  o.put("-loadreplay", string(""), "replay actions from file");
  o.put("-savereplay", string(""), "save actions to file (server only)");
  o.put("-savefile",   string(""), "file to save state to");
  o.put("-loadfile",   string(""), "file to load state from");

  o.put("-ids", string(""),        "accepted player id:key pairs separated by ,");  
}


class PlayerClient : public IO_Buffer
{
public:

  enum {
    STAT_WINDOW = 8,       //  I/O history window length
  };
  
  string name, host;
  sint4  port;

  bool good;                // true if all is fine

  // previous total (for computing #bytes in current frame)  
  sint8 prev_rcvd_bytes, prev_sent_bytes; 

  History<double> bytes_rcvd, bytes_sent, msgs_rcvd;

  sint4 next_action_id;      // id of oldest msg in receive buffer
  sint4 merged_action_msgs;  // how many action msgs were merged
  sint4 skipped_action_msgs; // how many action msgs were skipped  
  sint4 read_actions_msgs;   // number of msgs read
  sint4 last_action_id;      // id of last executed action msg
  
  PlayerClient(TCPsocket so, sint4 clev) :
    IO_Buffer(so, clev, 0),
    bytes_rcvd(STAT_WINDOW),
    bytes_sent(STAT_WINDOW),
    msgs_rcvd(STAT_WINDOW)    
  {
    good = true;
    prev_rcvd_bytes = prev_sent_bytes = 0;
    next_action_id = 0;
    last_action_id = -1; // no message used in first frame
    skipped_action_msgs = 0;
    merged_action_msgs = 0;
  }

  virtual ~PlayerClient() { }

  // #rcvd bytes since last check
  sint8 recently_rcvd_bytes() {
    sint8 p = prev_rcvd_bytes;
    prev_rcvd_bytes = rcvd_bytes();
    return prev_rcvd_bytes - p;
  }

  // #sent bytes since last check  
  sint8 recently_sent_bytes() {
    sint8 p = prev_sent_bytes;
    prev_sent_bytes = sent_bytes();
    return prev_sent_bytes - p;
  }
  
};



IO_Buffer *Server::new_client(TCPsocket so, sint4 clev) const
{
  return new PlayerClient(so, clev);
#if 0
  if (save == "")
    {
      printf("Creating player client\n");
      return new PlayerClient(so, clev);
    }
  else {
    printf("Creating logging player client\n");
    return new LoggingPlayerClient(so, clev, save);
  }
#endif
}


void Server::status(std::ostream &os) const
{
#if 1
  os << endl;
  form(os,
       "fr=%d @ %.1fp %.1fa, R[%.2f %.1f] W[%.2f %.1f]",
       frame,
       (1000.0/(frame_time.prev()+0.000001)),
       (1000.0/(frame_time.avg()+0.000001)),       
       bytes_rcvd.avg()*0.001,
       bytes_rcvd.sum()/(frame_time.sum()+0.001),
       bytes_sent.avg()*0.001,
       bytes_sent.sum()/(frame_time.sum()+0.001));
  os << endl;
#else
  form(os,
       "%d %f %f W[%.2f %.1f]",
       frame,
       bytes_rcvd.avg()*1.0,
       bytes_rcvd.sum()*1.0,
       bytes_sent.avg()*0.001,
       bytes_sent.sum()*1000.0/frame_time.sum());
  os << endl;
#endif
}


void Server::send_messages(const Vector<std::string> &msgs)
{
  if (msgs.size() != clients.size()) ERR("wrong player count");

  FORU (i, clients.size()) {
    PlayerClient &pc = *dynamic_cast<PlayerClient*>(clients[i]);
    if (pc.good) {
      if (pc.write(msgs[i]) < 0) pc.good = false;
    }
  }
}

void Server::save_messages(const Vector<string> &msgs)
{
  if (msgs.size() != clients.size()) ERR("wrong player count");

  uint4 val = msgs.size();

  bool ok = fwrite(&val, sizeof(val), 1, save_file) == 1;

  FORU (i, msgs.size()) {
    //printf("Saving message %d of %d, size %d\n", i+1, msgs.size(), msgs[i].size());
    val = msgs[i].size();
    ok &= fwrite(&val, sizeof(val), 1, save_file) == 1;
    //fprintf(save_file, "%u\n", msgs[i].size());
    ok &= fwrite(msgs[i].data(), sizeof(char), msgs[i].size(), save_file) == msgs[i].size();
  }

  ok &= (fflush(save_file) == 0);
  if (!ok) ERR("write to replay file failed 1");
}


void Server::load_messages(Vector<string> &msgs)
{
  if (clients.size() == 0) return;

  msgs.clear();
  msgs.resize(clients.size());
  uint4 num_msgs, msg_size;

  if (fread(&num_msgs, sizeof(num_msgs), 1, load_file) != 1) {
    if (feof(load_file)) {
      fclose(load_file); load = false; REM("Reached replay EOF\n"); return;
    }      
    ERR("replay file corrupt 1");
  }
  
  if (num_msgs != clients.size()) ERR("Illegal client size in saved file");

  FORU (i, num_msgs) {
    int result = fread(&msg_size, sizeof(msg_size), 1, load_file);

    //printf("Load message %d of %d, res=%d, size %d\n", i+1, num_msgs, result, msg_size);

    if ((result == 0) || (result == EOF)) {
      fclose(load_file); load = false; REM("Reached replay EOF\n"); break;
    }

    msgs[i].clear();

    if (msg_size > 0) {

      char *place = new char[msg_size];
      
      if (fread(place, sizeof(char), msg_size, load_file) != msg_size)
	ERR("replay file corrupt 2");
      
      if (feof(load_file)) {
	fclose(load_file); load = false; REM("Reached replay EOF\n"); break;
      }
      
      msgs[i].append(place, msg_size);
      delete [] place; place = 0;
    }
  }
}


void Server::recv_messages(Vector<string> &msgs)
{
  msgs.clear();
  msgs.resize(clients.size()); // empty strings

  FORU (i, clients.size()) {
    PlayerClient &pc = *dynamic_cast<PlayerClient*>(clients[i]);

    if (pc.good) {

      pc.skipped_action_msgs = pc.merged_action_msgs = 0;

      // merge messages in receive buffer

      // 1. skip messages with id < frame - GameConst::ACTION_EXPIRATION
      // 2. merge messages with id <= frame
      
      /*
	<= frame  < frame-GameConst::ACTION_EXPIRATION
	|-current-|
	|         |-expired-|
	|         |       next_action_id
	|         v       v
	|-|-|-|-|-|-|-|-|-|-|-|
	^                   ^
        new       >         old

      */

      // fixme: possible race when communication is faster than loop
      //        very unlikely ...

      // remove expired actions

      string msg;
      bool message_read = false;
      sint4 r, merged = 0;
      
      while (pc.next_action_id < frame-GameConst::ACTION_EXPIRATION) {
	FOREVER { // in synchro mode: block if no message has arrived
	  r = pc.read(msg);
	  if (r != 0 || !synchro) break;
	  SDL_Delay(10);
	}
	if (r < 0) { pc.good = false; break; } // read error
	if (r <= 0) break; // nothing here
	message_read = true;
	++pc.skipped_action_msgs; // skip message
	++pc.next_action_id;	  
      }
      
      if (!pc.good) { msgs[i].clear(); continue; } // read error

      // merge remaining actions with id <= frame

      msgs[i].clear();

      while (pc.next_action_id <= frame) {
	FOREVER { // in synchro mode: block if no message has arrived
	  r = pc.read(msg);
	  if (r != 0 || !synchro) break;
	  SDL_Delay(10);	  
	}
	if (r < 0) { pc.good = false; break; } // read error
	if (r <= 0) break; // nothing here
	message_read = true;	
	++merged;
	++pc.next_action_id;
	msgs[i] += msg;
      }

#if 0      
      cout << "-------------------" << endl;
      cout << msg << endl;
#endif
      
      if (!pc.good) { msgs[i].clear(); continue; } // read error

      if (!message_read) pc.last_action_id = -1; // no message
      else { // at least one message
	pc.last_action_id = pc.next_action_id;
      }

      pc.merged_action_msgs = merged;
      pc.read_actions_msgs = pc.skipped_action_msgs + merged;

#if 0
      // debug output
      if (pc.skipped_action_msgs > 0 || !message_read || merged > 1) {
	errstr << "REMARK: player " << i;
	if (pc.skipped_action_msgs > 0)
	  errstr << " skipped=" << pc.skipped_action_msgs;
	if (!message_read)
	  errstr << " no action in buffer";
	if (merged > 1)
	  errstr << " merged=" << merged;
	errstr << endl;
      }
#endif      
    }
  }
}

Server::Server() :
  TCP_Server(),
  bytes_sent(STAT_WINDOW),
  bytes_rcvd(STAT_WINDOW),
  frame_time(STAT_WINDOW)
{
  synchro = false;
  game = 0;
  frame = 0;
  save = load = 0;
  save_file = load_file = 0;
  start_time = 0;
}


Server::~Server()
{
  FORALL (clients, i) {
    TCPsocket t = (*i)->get_socket();
    SDLNet_TCP_Close(t);
    delete *i;
  }

  clients.clear();

  if (game) { delete game; game = 0; }
}


void Server::run(const string &world, Server::Options &opt)
{
  Game::Options go;

  set_clev(opt.clev);
  synchro = opt.synchro;

  string server_world = world;
  string file_name = opt.load_replay;

  if (file_name == "") load = false;
  else {
    file_name = sanitize_path(file_name);
    load = ((load_file = fopen(file_name.c_str(), "r")) != 0);
    if (!load) ERR2("could not open replay file", file_name);
    
    server_world.clear();
    uint4 size;
    if (fread(&size, sizeof(size), 1, load_file) != 1)
      ERR("insufficient data in replay file");

    assert(size > 0);
    char *loaded_world = new char[size];
    if (fread(loaded_world, sizeof(*loaded_world), size, load_file) != size)
      ERR("Did not read correct world size.");

    server_world.append(loaded_world, size);
    delete [] loaded_world; loaded_world = 0;
  }

  file_name = opt.save_replay;
  if (file_name == "") save = false;
  else {
    file_name = sanitize_path(file_name);
    save = ((save_file = fopen(file_name.c_str(), "w+")) != 0);
    if (!save) ERR2("could not open replay file", file_name); 
    uint4 val = server_world.size();
    if (fwrite(&val, sizeof(val), 1, save_file) != 1 ||
	fwrite(server_world.data(), sizeof(char), server_world.size(), save_file) !=
	server_world.size()) ERR("write to replay file failed 2");
  }

  go.save_file = opt.save_file;
  go.load_file = opt.load_file;

  game = new Game(server_world, go);
  string client_world = game->get_client_world();

  frame = 0;

  // check ids

  Vector<string> id_strs;

  if (opt.ids.empty()) {

    id_strs.push_back("");

  } else {

    tokenize(opt.ids, id_strs, ",");
    if ((int)id_strs.size() != game->get_player_num()) {
      ERR("#ids does not match #players: " << opt.ids << " " <<  game->get_player_num());
    }

    Vector<string> sorted_ids = id_strs;

    sort(sorted_ids.begin(), sorted_ids.end());

    FORS (i, game->get_player_num()-1) {
      if (sorted_ids[i] == sorted_ids[i+1])
	ERR("ids must be distinct: " << opt.ids);
    }
  }

  // wait for all players to connect

  init_listen(opt.port);
  
  cout << "Waiting for " << game->get_player_num()
       << " player(s) to connect to " << opt.port
       << " ... " << flush;

  if (opt.freq <= 0) ERR("freq <= 0");

  call_handlers(Event(FROM, READY_MSG));

  FOREVER {

    if (shut_down) return;

    // fixme: handle disconnects here, too!

    // thoang: add set message to the worker
	
    if (handle_connection()) {

      cout << "connected: client " << clients.size()-1 << " ";
      
      // get the client's IP and port number

      const IPaddress *remoteip = SDLNet_TCP_GetPeerAddress(clients.back()->get_socket());

      if (!remoteip) {

	REM("SDLNet_TCP_GetPeerAddress error: " << SDLNet_GetError() << " ");

      } else {
	
	uint4 ipaddr = SDL_SwapBE32(remoteip->host);

	cout <<
	  boost::format("from %d.%d.%d.%d port %u\n") %
	  (ipaddr >> 24) % 
	  ((ipaddr >> 16) & 0xff) %
	  ((ipaddr >> 8) & 0xff) %
	  (ipaddr & 0xff) %
	  remoteip->port
	     << endl;
      }
      
      if (clients.size() == uint4(game->get_player_num())) break;
    }
    SDL_Delay(10);
  }

  close(); // close server socket

  cout << "OK. Now waiting for authentication ..." << flush;
    
  uint4 start = SDL_GetTicks();
  Vector<sint1> ok(clients.size(), 0);
  Vector<string> auth_strs(clients.size());
  uint4 ok_n = 0;

  while (ok_n != clients.size() && SDL_GetTicks() - start < AUTH_WAIT_SEC*1000) {

    FORU (i, clients.size()) {
	
      PlayerClient &pc = *dynamic_cast<PlayerClient*>(clients[i]);
	
      if (pc.good) {
	  
	string msg;
	sint4 r = pc.read(msg);
	  
	if (r == 0) continue;
	if (r < 0) { pc.good = false; continue; } // read error
	  
	cout << "read: (" << i << ",'" << msg << "') " << endl;

	auth_strs[i] = msg;
	
	FIND (id_strs, j, msg);
	ok[i] = (j != id_strs.end());

	if (ok[i]) ++ok_n;
	else {
	  REM("wrong authentication! " << i << " " << msg);
	}
      }
    }

    SDL_Delay(10);
  }

  if (ok_n != clients.size()) {
    ERR("" << (clients.size()-ok_n) << " player(s) not authenticated!");
  }

  // assign client ids

  cout << "assign client ids ... ";
  
  if (opt.ids.empty()) {

    // shuffle if no authentication is used
    random_shuffle(clients.begin(), clients.end());

    cout << "shuffle" << endl;
    
  } else {

    // authentication strings must be distinct
  
    Vector<string> sorted_auth_strs = auth_strs;
    sort(sorted_auth_strs.begin(), sorted_auth_strs.end());
    
    FORS (i, clients.size()-1) {
      if (sorted_auth_strs[i] == sorted_auth_strs[i+1])
	ERR("duplicate authentication strings! " << sorted_auth_strs[i]);
    }
  
    // assignment based on authentication strings

    Vector<IO_Buffer*> new_clients;
    
    FORU (i, clients.size()) {
      FIND(id_strs, j, auth_strs[i]);
      if (j == id_strs.end()) ERR("id not found");

      cout << "[" << i << "<-" << j-id_strs.begin() << "] ";
      new_clients.push_back(clients[j-id_strs.begin()]);
    }

    cout << endl;
    clients = new_clients;
  }

  cout << "OK. Send game information ... " << flush;

  Vector<string> views, actions;
  FORS (i, game->get_player_num()) { views.push_back(""); actions.push_back(""); }

  FORS (i, game->get_player_num()) {
    stringstream ss;
    ss << i << endl << client_world;  // client player id + public world description
    views[i] = ss.str();
  }

  send_messages(views);
  
  cout << "OK. Run the game ..." << endl;

  // run the game

  sint4 ft = my_round(1000/opt.freq); // frame time (ms)

  FOREVER {

    if (shut_down || frame == opt.tmax) break;

    sint4 start = SDL_GetTicks();
    
    game->preprocess_actions(actions);

    if (!game->get_save_req()) {
      // one simulation step
      game->simulation_step(actions, views);
      call_handlers(Event(FROM, STEP_MSG));
    } else {
      cout << "SAVE REQUEST" << endl;
    }
    
    // prepend frame information
    
    FORALL (clients, i) {
      PlayerClient &pc = *dynamic_cast<PlayerClient*>(*i);

      string u;

      // also change corresponding code in View2.C
      
      sint4 a = frame;                          // current frame
      sint4 b = pc.last_action_id;              // action executed in this frame
      sint4 c = pc.skipped_action_msgs;         // -1, no action deleted
      sint4 d = pc.merged_action_msgs;
      sint4 e = game->get_save_req() ? 1 : 0;   // message flag: save game
      sint4 f = my_round(opt.freq * 1024);      // send simulation rate
	
      write_b_int4(u, a);
      write_b_int4(u, b);
      write_b_int4(u, c);
      write_b_int4(u, d);
      write_b_int4(u, e);
      write_b_int4(u, f);      

      views[i-clients.begin()] = u + views[i-clients.begin()];
    }

    send_messages(views); // send views

    if (game->get_save_req()) {
      ostringstream oss;
      game->save(oss);
      string file = game->get_savefile();
      ofstream savefile(file.c_str(), ios::binary);
      savefile.write(oss.str().c_str(), oss.str().size());
      savefile.close();
      cout << "SAVING TO " << file << " " << oss.str().size() << endl;

      SDL_Delay(1000);
      break;
    }

    static bool waited = false;
    
    if (!waited && opt.sdelay > 0) {
      waited = true;
      cout << "Wait " << opt.sdelay << " seconds" << endl;
      SDL_Delay(opt.sdelay*1000);
    }
  
    call_handlers(Event(FROM, STEP_MSG));
    
    // fixme: work around the stupid 10ms Linux timer resolution by
    // averaging over multiple frames (will be resolved in kernel 2.6.x)
    
    sint4 diff = ft - (SDL_GetTicks() - start);

    if (diff > 0) {
      // cout << "delay " << diff << endl;
      SDL_Delay(diff);
    }

    recv_messages(actions); // need them out of the queue no matter what
    if (load) load_messages(actions); // overwrite them from a file
    if (save) save_messages(actions); // save them to file
    
    // update time statistics

    sint4 sent=0, rcvd=0;
    
    FORALL (clients, j) {
      
      PlayerClient &pc = *dynamic_cast<PlayerClient*>(*j);

      sint8 r_sent = pc.recently_sent_bytes();
      sint8 r_rcvd = pc.recently_rcvd_bytes();
      
      sent += r_sent;
      rcvd += r_rcvd;

      pc.bytes_sent.append(r_sent);
      pc.bytes_rcvd.append(r_rcvd);

      pc.msgs_rcvd.append(pc.read_actions_msgs);
    }
    
    bytes_sent.append(sent);
    bytes_rcvd.append(rcvd);

    if (opt.act) game->gen_actions(actions);    

    if (game->get_finished()) break;

    frame_time.append(SDL_GetTicks() - start);
    frame++;
    //    cout << frame << " " << sent << " " << rcvd << endl;

    if ((frame % 16) == 0) {
      
      status(cout);
      bool alive_client = false;
      
      FORALL (clients, i) {
	
	PlayerClient &pc = *dynamic_cast<PlayerClient*>(*i);

	// tmanager recognizes "dead" messages:  ^<id>: dead
	// don't change that!
	
	cout << i-clients.begin() << ": ";
	
	if (pc.good) {

          alive_client = true;

	  form(cout, "RB[%.2fp %.2fa] WB[%.2fp %.2fa] RM[%.2fa] I[%dp] O[%dp]",
	       pc.bytes_rcvd.prev()*0.001, pc.bytes_rcvd.avg()*0.001,
	       pc.bytes_sent.prev()*0.001, pc.bytes_sent.avg()*0.001,
	       pc.msgs_rcvd.avg(),
	       pc.read_buf_len(), pc.write_buf_len()
	       );
	  cout << endl;
	} else {
	  
	  cout << " dead" << endl;
	}
      }
      if (!alive_client) {
        if (opt.keepalive) {
          ft = 0;
        } else {
          break;
        }
      }
    }
  
    if (opt.wait < 0) {
      char s[100]; fgets(s, 80, stdin); // wait for <enter>
    } else if (opt.wait > 0) {
      SDL_Delay(opt.wait);                  // wait time
    }

  }
  
  if (load) fclose(load_file);
  if (save) fclose(save_file);

  // keep clients and game alive to send out result
}



#if 0
class LoggingPlayerClient : public PlayerClient {
public:
  LoggingPlayerClient(TCPsocket so, sint4 clev, string fname) :
    PlayerClient(so, clev)
  {
    if ((fin = fopen((fname+".in").c_str(), "w+")) == 0)
      { fin = 0; REM("Input open failure"); }
    if ((fout = fopen((fname+".out").c_str(), "w+")) == 0)
      { fout = 0; REM("Output open failure"); }
  }
  ~LoggingPlayerClient()
  { if (fin) fclose(fin); if (fout) fclose(fout); }

  sint4 write(const string &s)
  {
    if (fout) {
      uint4 val = s.size();
      fwrite(&val, sizeof(val), 1, fout);
      //fprintf(fout, "%d\n", s.size()); // s.c_str()
      fwrite(s.data(), sizeof(char), s.size(), fout);
    }
    return PlayerClient::write(s);
  }

  sint4 read(string &s)
  {
    sint4 res = PlayerClient::read(s);
    if (fin) {
      int val = s.size();
      fwrite(&val, sizeof(val), 1, fout);
      //fprintf(fin, "%d\n", s.size()); // s.c_str()
      fwrite(s.data(), sizeof(char), s.size(), fin);
    }
    return res;
  }

private:
  FILE *fin, *fout;
};
#endif

