// $Id: Quadtree.C,v 1.1.1.1 2004/10/11 09:09:19 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

// Quadtree implementation
// generates region adjecency graph on the fly 

#include "Quadtree.H"
#include "SDLinit.H"
#include "SDL_gfxPrimitives.h"
#include "GUI.H"

const sint4 GFX_MAG = 1; // magnification

static bool gfx = true;
static uint4 updates, splits, edges;

QuadtreeNode::QuadtreeNode()
{
  pred = succ[0] = succ[1] = succ[2] = succ[3] = 0;
  xl = yu = w = 0;
}

QuadtreeNode::~QuadtreeNode()
{
  FORS (i, 4) if (succ[i]) delete succ[i];
}


const QuadtreeNode *QuadtreeNode::locate(sint4 x, sint4 y) const
{
  if (x < xl || x >= xl+w || y < yu || y >= yu+w) ERR("point outside square");

  if (!is_split()) return this;

  sint4 sx = xl + (w >> 1);
  sint4 sy = yu + (w >> 1);
  
  if (x < sx) {
    
    if (y < sy) return succ[S_UL]->locate(x, y); else return succ[S_LL]->locate(x, y);
    
  } else {
    
    if (y < sy) return succ[S_UR]->locate(x, y); else return succ[S_LR]->locate(x, y);
  }
}


bool QuadtreeNode::insert(sint4 x, sint4 y)
{
  if (x < xl || x >= xl+w || y < yu || y >= yu+w) {
    cerr << x << " " << y << " | " << xl << " " << yu << " " << w << endl;
    ERR("point outside square");
  }

  if (w > 1) {
  
    if (!is_split()) {
      split();
      if (w == 2) return true;
    }

    sint4 sx = xl + (w >> 1);
    sint4 sy = yu + (w >> 1);
    
    if (x < sx) {
      
      if (y < sy) return succ[S_UL]->insert(x, y); else return succ[S_LL]->insert(x, y);
      
    } else {
      
      if (y < sy) return succ[S_UR]->insert(x, y); else return succ[S_LR]->insert(x, y);
    }
  }

  return false;
}


// connect child with neighbor

inline static void connect(QuadtreeNode &parent, QuadtreeNode &from,
			   sint4 t, sint4 fldir, sint4 tldir)
{
  QuadtreeNode &to = *parent.succ[t]; 
  QuadtreeNode::NList &fl = from.nls[fldir];      // from-list
  QuadtreeNode::NList &tl = to.nls[tldir];        // to-list
  QuadtreeNode::Edge e;

  ++edges;
  assert(tl.empty() && fl.empty());
  
  tl.push_back(e);
  fl.push_back(e);
  tl.begin()->node     = &from;
  tl.begin()->list_pos = fl.begin();
  fl.begin()->node     = &to;
  fl.begin()->list_pos = tl.begin();
}


const sint4 HORI = 0, VERT = 1; // how children are aligned

void update_edges(QuadtreeNode *node,
		  sint4 n_dir, sint4 n_dir_c, // dir to neighbor, complement
		  sint4 align,                // HORI or VERT alignment of successors
		  sint4 succ1, sint4 succ2    // successor indexes (West/East North/S)
		  )
{    
  const sint4 w2 = node->w >> 1;  // width of children
  const sint4 sx = node->xl + w2; // midpoint
  const sint4 sy = node->yu + w2;  

  FORALL (node->nls[n_dir], i) {              // traverse through neighbor list

    ++updates;
    QuadtreeNode &n = *(*i).node;              // neighbor is n
    QuadtreeNode::NList &nl = n.nls[n_dir_c];  // neighbor list of n

    bool half[2] = { n.xl < sx, n.yu < sy };   // prepare data-driven comparison below
    // 0: true:West  false:East
    // 1: true:North false:South

    if (n.w < node->w) {

      // neighbor is smaller -> only one connection 
	
      assert(nl.size() == 1);

      // choose child to connect to
      QuadtreeNode *child = half[align] ? node->succ[succ1] : node->succ[succ2];
      QuadtreeNode::Edge e;
      QuadtreeNode::NList &cnl = child->nls[n_dir]; // child neighbor list

      // new entry in child n-list
      QuadtreeNode::NList::iterator lp = cnl.insert(cnl.end(), e);
	
      assert(nl.begin() == (*i).list_pos); // one element

      // connect neighbor to child
      nl.begin()->list_pos = lp; 
      nl.begin()->node = child;

      // connect child to neighbor
      lp->node = &n;
      lp->list_pos = (*i).list_pos;
	
    } else {

      // neighbor covers entire square -> link both children to neighbor

      QuadtreeNode::Edge e;
      QuadtreeNode *child1 = node->succ[succ1];
      QuadtreeNode *child2 = node->succ[succ2];	
      QuadtreeNode::NList &cnl1 = child1->nls[n_dir]; // child1 neighbor list
      QuadtreeNode::NList &cn